import pytest

from ..state import State
from ..helpers_state import PydraStateError


@pytest.mark.parametrize(
    "inputs, splitter, ndim, states_ind, states_val, group_for_inputs, " "groups_stack",
    [
        (
            {"NA.a": [3, 5]},
            "a",
            1,
            [{"NA.a": 0}, {"NA.a": 1}],
            [{"NA.a": 3}, {"NA.a": 5}],
            {"NA.a": 0},
            [[0]],
        ),
        (
            {"NA.a": [3, 5], "NA.b": ["str1", "str2"]},
            ("a", "b"),
            1,
            [{"NA.a": 0, "NA.b": 0}, {"NA.a": 1, "NA.b": 1}],
            [{"NA.a": 3, "NA.b": "str1"}, {"NA.a": 5, "NA.b": "str2"}],
            {"NA.a": 0, "NA.b": 0},
            [[0]],
        ),
        (
            {"NA.a": [3, 5], "NA.b": ["str1", "str2"]},
            ["a", "b"],
            2,
            [
                {"NA.a": 0, "NA.b": 0},
                {"NA.a": 0, "NA.b": 1},
                {"NA.a": 1, "NA.b": 0},
                {"NA.a": 1, "NA.b": 1},
            ],
            [
                {"NA.a": 3, "NA.b": "str1"},
                {"NA.a": 3, "NA.b": "str2"},
                {"NA.a": 5, "NA.b": "str1"},
                {"NA.a": 5, "NA.b": "str2"},
            ],
            {"NA.a": 0, "NA.b": 1},
            [[0, 1]],
        ),
        (
            {"NA.a": [3, 5], "NA.b": "str1"},
            "a",
            1,
            [{"NA.a": 0}, {"NA.a": 1}],
            [{"NA.a": 3}, {"NA.a": 5}],
            {"NA.a": 0},
            [[0]],
        ),
        (
            {"NA.a": [3, 5], "NA.b": ["str1", "str2"], "NA.c": [10, 20]},
            [("a", "c"), "b"],
            2,
            [
                {"NA.a": 0, "NA.b": 0, "NA.c": 0},
                {"NA.a": 0, "NA.b": 1, "NA.c": 0},
                {"NA.a": 1, "NA.b": 0, "NA.c": 1},
                {"NA.a": 1, "NA.b": 1, "NA.c": 1},
            ],
            [
                {"NA.a": 3, "NA.b": "str1", "NA.c": 10},
                {"NA.a": 3, "NA.b": "str2", "NA.c": 10},
                {"NA.a": 5, "NA.b": "str1", "NA.c": 20},
                {"NA.a": 5, "NA.b": "str2", "NA.c": 20},
            ],
            {"NA.a": 0, "NA.c": 0, "NA.b": 1},
            [[0, 1]],
        ),
    ],
)
def test_state_1(
    inputs, splitter, ndim, states_ind, states_val, group_for_inputs, groups_stack
):
    """ single state: testing groups, prepare_states and prepare_inputs"""
    st = State(name="NA", splitter=splitter)
    assert st.splitter == st.right_splitter
    assert st.splitter_rpn == st.right_splitter_rpn
    assert st.left_splitter is None
    assert st.left_combiner_all == []

    st.prepare_states(inputs)
    assert st.group_for_inputs_final == group_for_inputs
    assert st.groups_stack_final == groups_stack

    assert st.states_ind == states_ind
    assert st.states_val == states_val

    st.prepare_inputs()
    assert st.inputs_ind == states_ind


def test_state_2_err():
    with pytest.raises(PydraStateError) as exinfo:
        st = State("NA", splitter={"a"})
    assert "splitter has to be a string, a tuple or a list" == str(exinfo.value)


def test_state_3_err():
    with pytest.raises(PydraStateError) as exinfo:
        st = State("NA", splitter=["a", "b"], combiner=("a", "b"))
    assert "combiner has to be a string or a list" == str(exinfo.value)


def test_state_4_err():
    st = State("NA", splitter="a", combiner=["a", "b"])
    with pytest.raises(PydraStateError) as exinfo:
        st.combiner_validation()
    assert "all combiners have to be in the splitter" in str(exinfo.value)


def test_state_5_err():
    st = State("NA", combiner="a")
    with pytest.raises(PydraStateError) as exinfo:
        st.combiner_validation()
    assert "splitter has to be set before" in str(exinfo.value)


def test_state_connect_1():
    """ two 'connected' states: testing groups, prepare_states and prepare_inputs
        no explicit splitter for the second state
    """
    st1 = State(name="NA", splitter="a")
    st2 = State(name="NB", other_states={"NA": (st1, "b")})
    assert st2.splitter == "_NA"
    assert st2.splitter_rpn == ["NA.a"]
    assert st2.left_splitter == st2.splitter
    assert st2.left_splitter_rpn == st2.splitter_rpn
    assert st2.right_splitter is None
    assert st2.right_splitter_rpn == []

    st2.prepare_states(inputs={"NA.a": [3, 5]})
    assert st2.group_for_inputs_final == {"NA.a": 0}
    assert st2.groups_stack_final == [[0]]
    assert st2.states_ind == [{"NA.a": 0}, {"NA.a": 1}]
    assert st2.states_val == [{"NA.a": 3}, {"NA.a": 5}]

    st2.prepare_inputs()
    assert st2.inputs_ind == [{"NB.b": 0}, {"NB.b": 1}]


def test_state_connect_1a():
    """ two 'connected' states: testing groups, prepare_states and prepare_inputs
        the second state has explicit splitter from the first one (Left part)
    """
    st1 = State(name="NA", splitter="a")
    st2 = State(name="NB", splitter="_NA", other_states={"NA": (st1, "b")})
    assert st2.splitter == "_NA"
    assert st2.splitter_rpn == ["NA.a"]

    st2.prepare_states(inputs={"NA.a": [3, 5]})
    assert st2.group_for_inputs_final == {"NA.a": 0}
    assert st2.groups_stack_final == [[0]]
    assert st2.states_ind == [{"NA.a": 0}, {"NA.a": 1}]
    assert st2.states_val == [{"NA.a": 3}, {"NA.a": 5}]

    st2.prepare_inputs()
    assert st2.inputs_ind == [{"NB.b": 0}, {"NB.b": 1}]


def test_state_connect_1b_exception():
    """can't provide explicitly NA.a (should be _NA)"""
    st1 = State(name="NA", splitter="a", other_states={})
    st2 = State(name="NB", splitter="NA.a")
    with pytest.raises(PydraStateError) as excinfo:
        st2.splitter_validation()
    assert "consider using _NA" in str(excinfo.value)


@pytest.mark.parametrize("splitter2, other_states2", [("_NA", {}), ("_N", {"NA": ()})])
def test_state_connect_1c_exception(splitter2, other_states2):
    """can't ask for splitter from node that is not connected"""
    st1 = State(name="NA", splitter="a")
    with pytest.raises(PydraStateError) as excinfo:
        st2 = State(name="NB", splitter=splitter2, other_states=other_states2)
        st2.splitter_validation()


def test_state_connect_2():
    """ two 'connected' states: testing groups, prepare_states and prepare_inputs
         the second state has explicit splitter that contains
         splitter from the first node and a new field (Left and Right parts)
    """
    st1 = State(name="NA", splitter="a")
    st2 = State(name="NB", splitter=["_NA", "a"], other_states={"NA": (st1, "b")})

    assert st2.splitter == ["_NA", "NB.a"]
    assert st2.splitter_rpn == ["NA.a", "NB.a", "*"]
    assert st2.left_splitter == "_NA"
    assert st2.left_splitter_rpn == ["NA.a"]
    assert st2.right_splitter == "NB.a"
    assert st2.right_splitter_rpn == ["NB.a"]

    st2.update_connections()
    st2.prepare_states(inputs={"NA.a": [3, 5], "NB.a": [1, 2]})
    assert st2.group_for_inputs_final == {"NA.a": 0, "NB.a": 1}
    assert st2.groups_stack_final == [[0, 1]]
    assert st2.keys_final == ["NA.a", "NB.a"]

    assert st2.states_ind == [
        {"NA.a": 0, "NB.a": 0},
        {"NA.a": 0, "NB.a": 1},
        {"NA.a": 1, "NB.a": 0},
        {"NA.a": 1, "NB.a": 1},
    ]
    assert st2.states_val == [
        {"NA.a": 3, "NB.a": 1},
        {"NA.a": 3, "NB.a": 2},
        {"NA.a": 5, "NB.a": 1},
        {"NA.a": 5, "NB.a": 2},
    ]

    st2.prepare_inputs()
    assert st2.inputs_ind == [
        {"NB.b": 0, "NB.a": 0},
        {"NB.b": 0, "NB.a": 1},
        {"NB.b": 1, "NB.a": 0},
        {"NB.b": 1, "NB.a": 1},
    ]


def test_state_connect_2a():
    """ two 'connected' states: testing groups, prepare_states and prepare_inputs
         the second state has explicit splitter that contains
         splitter from the first node and a new field;
         adding an additional scalar field that is not part of the splitter
    """
    st1 = State(name="NA", splitter="a")
    st2 = State(name="NB", splitter=["_NA", "a"], other_states={"NA": (st1, "b")})

    assert st2.splitter == ["_NA", "NB.a"]
    assert st2.splitter_rpn == ["NA.a", "NB.a", "*"]

    st2.prepare_states(inputs={"NA.a": [3, 5], "NB.a": [1, 2], "NB.s": 1})
    assert st2.group_for_inputs_final == {"NA.a": 0, "NB.a": 1}
    assert st2.groups_stack_final == [[0, 1]]
    assert st2.keys_final == ["NA.a", "NB.a"]

    assert st2.states_ind == [
        {"NA.a": 0, "NB.a": 0},
        {"NA.a": 0, "NB.a": 1},
        {"NA.a": 1, "NB.a": 0},
        {"NA.a": 1, "NB.a": 1},
    ]
    assert st2.states_val == [
        {"NA.a": 3, "NB.a": 1},
        {"NA.a": 3, "NB.a": 2},
        {"NA.a": 5, "NB.a": 1},
        {"NA.a": 5, "NB.a": 2},
    ]

    st2.prepare_inputs()
    assert st2.inputs_ind == [
        {"NB.b": 0, "NB.a": 0},
        {"NB.b": 0, "NB.a": 1},
        {"NB.b": 1, "NB.a": 0},
        {"NB.b": 1, "NB.a": 1},
    ]


def test_state_connect_2b():
    """ two 'connected' states: testing groups, prepare_states and prepare_inputs
        the second state has explicit splitter with a new field (Right part)
        splitter from the first node (Left part) has to be added
    """
    st1 = State(name="NA", splitter="a")
    st2 = State(name="NB", splitter="a", other_states={"NA": (st1, "b")})

    assert st2.splitter == ["_NA", "NB.a"]
    assert st2.splitter_rpn == ["NA.a", "NB.a", "*"]
    assert st2.right_splitter == "NB.a"
    assert st2.left_splitter == "_NA"

    st2.prepare_states(inputs={"NA.a": [3, 5], "NB.a": [1, 2]})
    assert st2.group_for_inputs_final == {"NA.a": 0, "NB.a": 1}
    assert st2.groups_stack_final == [[0, 1]]
    assert st2.states_ind == [
        {"NA.a": 0, "NB.a": 0},
        {"NA.a": 0, "NB.a": 1},
        {"NA.a": 1, "NB.a": 0},
        {"NA.a": 1, "NB.a": 1},
    ]
    assert st2.states_val == [
        {"NA.a": 3, "NB.a": 1},
        {"NA.a": 3, "NB.a": 2},
        {"NA.a": 5, "NB.a": 1},
        {"NA.a": 5, "NB.a": 2},
    ]

    st2.prepare_inputs()
    assert st2.inputs_ind == [
        {"NB.b": 0, "NB.a": 0},
        {"NB.b": 0, "NB.a": 1},
        {"NB.b": 1, "NB.a": 0},
        {"NB.b": 1, "NB.a": 1},
    ]


def test_state_connect_3():
    """ three 'connected' states: testing groups, prepare_states and prepare_inputs
        the third state connected to two previous states;
        splitter from the previous states (Left part) has to be added
    """
    st1 = State(name="NA", splitter="a")
    st2 = State(name="NB", splitter="a")
    st3 = State(name="NC", other_states={"NA": (st1, "b"), "NB": (st2, "c")})

    assert st3.splitter == ["_NA", "_NB"]
    assert st3.splitter_rpn == ["NA.a", "NB.a", "*"]
    assert st3.left_splitter == st3.splitter
    assert st3.left_splitter_rpn == st3.splitter_rpn
    assert st3.left_splitter_rpn_compact == ["_NA", "_NB", "*"]
    assert st3.right_splitter is None
    assert st3.right_splitter_rpn == []

    st3.prepare_states(inputs={"NA.a": [3, 5], "NB.a": [30, 50]})
    assert st3.group_for_inputs_final == {"NA.a": 0, "NB.a": 1}
    assert st3.groups_stack_final == [[0, 1]]
    assert st3.states_ind == [
        {"NA.a": 0, "NB.a": 0},
        {"NA.a": 0, "NB.a": 1},
        {"NA.a": 1, "NB.a": 0},
        {"NA.a": 1, "NB.a": 1},
    ]
    assert st3.states_val == [
        {"NA.a": 3, "NB.a": 30},
        {"NA.a": 3, "NB.a": 50},
        {"NA.a": 5, "NB.a": 30},
        {"NA.a": 5, "NB.a": 50},
    ]

    st3.prepare_inputs()
    assert st3.inputs_ind == [
        {"NC.b": 0, "NC.c": 0},
        {"NC.b": 0, "NC.c": 1},
        {"NC.b": 1, "NC.c": 0},
        {"NC.b": 1, "NC.c": 1},
    ]


def test_state_connect_3a():
    """ three 'connected' states: testing groups, prepare_states and prepare_inputs
        the third state connected to two previous states;
        the third state has explicit splitter that contains splitters from previous states
    """
    st1 = State(name="NA", splitter="a")
    st2 = State(name="NB", splitter="a")
    st3 = State(
        name="NC",
        splitter=["_NA", "_NB"],
        other_states={"NA": (st1, "b"), "NB": (st2, "c")},
    )

    assert st3.splitter == ["_NA", "_NB"]
    assert st3.splitter_rpn == ["NA.a", "NB.a", "*"]

    st3.prepare_states(inputs={"NA.a": [3, 5], "NB.a": [30, 50]})
    assert st3.group_for_inputs_final == {"NA.a": 0, "NB.a": 1}
    assert st3.groups_stack_final == [[0, 1]]
    assert st3.states_ind == [
        {"NA.a": 0, "NB.a": 0},
        {"NA.a": 0, "NB.a": 1},
        {"NA.a": 1, "NB.a": 0},
        {"NA.a": 1, "NB.a": 1},
    ]
    assert st3.states_val == [
        {"NA.a": 3, "NB.a": 30},
        {"NA.a": 3, "NB.a": 50},
        {"NA.a": 5, "NB.a": 30},
        {"NA.a": 5, "NB.a": 50},
    ]

    st3.prepare_inputs()
    assert st3.inputs_ind == [
        {"NC.b": 0, "NC.c": 0},
        {"NC.b": 0, "NC.c": 1},
        {"NC.b": 1, "NC.c": 0},
        {"NC.b": 1, "NC.c": 1},
    ]


def test_state_connect_3b():
    """ three 'connected' states: testing groups, prepare_states and prepare_inputs
        the third state connected to two previous states;
        the third state has explicit splitter that contains splitter only from the first state.
        splitter from the second state has to be added (partial Left part)
    """
    st1 = State(name="NA", splitter="a")
    st2 = State(name="NB", splitter="a")
    st3 = State(
        name="NC", splitter="_NB", other_states={"NA": (st1, "b"), "NB": (st2, "c")}
    )

    assert st3.splitter == ["_NA", "_NB"]
    assert st3.splitter_rpn == ["NA.a", "NB.a", "*"]

    st3.prepare_states(inputs={"NA.a": [3, 5], "NB.a": [30, 50]})
    assert st3.group_for_inputs_final == {"NA.a": 0, "NB.a": 1}
    assert st3.groups_stack_final == [[0, 1]]
    assert st3.states_ind == [
        {"NA.a": 0, "NB.a": 0},
        {"NA.a": 0, "NB.a": 1},
        {"NA.a": 1, "NB.a": 0},
        {"NA.a": 1, "NB.a": 1},
    ]
    assert st3.states_val == [
        {"NA.a": 3, "NB.a": 30},
        {"NA.a": 3, "NB.a": 50},
        {"NA.a": 5, "NB.a": 30},
        {"NA.a": 5, "NB.a": 50},
    ]

    st3.prepare_inputs()
    assert st3.inputs_ind == [
        {"NC.b": 0, "NC.c": 0},
        {"NC.b": 0, "NC.c": 1},
        {"NC.b": 1, "NC.c": 0},
        {"NC.b": 1, "NC.c": 1},
    ]


def test_state_connect_4():
    """ three 'connected' states: testing groups, prepare_states and prepare_inputs
        the third state connected to two previous states;
        the third state has explicit scalar(!) splitter that contains two previous states
    """
    st1 = State(name="NA", splitter="a")
    st2 = State(name="NB", splitter="a")
    st3 = State(
        name="NC",
        splitter=("_NA", "_NB"),
        other_states={"NA": (st1, "b"), "NB": (st2, "c")},
    )

    assert st3.splitter == ("_NA", "_NB")
    assert st3.splitter_rpn == ["NA.a", "NB.a", "."]
    assert st3.left_splitter == st3.splitter
    assert st3.left_splitter_rpn == st3.splitter_rpn
    assert st3.left_splitter_rpn_compact == ["_NA", "_NB", "."]
    assert st3.right_splitter is None
    assert st3.right_splitter_rpn == []

    st3.prepare_states(inputs={"NA.a": [3, 5], "NB.a": [30, 50]})
    assert st3.group_for_inputs_final == {"NA.a": 0, "NB.a": 0}
    assert st3.groups_stack_final == [[0]]
    assert st3.states_ind == [{"NA.a": 0, "NB.a": 0}, {"NA.a": 1, "NB.a": 1}]
    assert st3.states_val == [{"NA.a": 3, "NB.a": 30}, {"NA.a": 5, "NB.a": 50}]

    st3.prepare_inputs()
    assert st3.inputs_ind == [{"NC.b": 0, "NC.c": 0}, {"NC.b": 1, "NC.c": 1}]


def test_state_connect_5():
    """ two 'connected' states: testing groups, prepare_states and prepare_inputs,
        the first state has outer splitter,
        the second state has no explicit splitter
    """
    st1 = State(name="NA", splitter=["a", "b"])
    st2 = State(name="NB", other_states={"NA": (st1, "a")})
    assert st2.splitter == "_NA"
    assert st2.splitter_rpn == ["NA.a", "NA.b", "*"]

    st2.prepare_states(inputs={"NA.a": [3, 5], "NA.b": [10, 20]})
    assert st2.group_for_inputs_final == {"NA.a": 0, "NA.b": 1}
    assert st2.groups_stack_final == [[0, 1]]
    assert st2.states_ind == [
        {"NA.a": 0, "NA.b": 0},
        {"NA.a": 0, "NA.b": 1},
        {"NA.a": 1, "NA.b": 0},
        {"NA.a": 1, "NA.b": 1},
    ]
    assert st2.states_val == [
        {"NA.a": 3, "NA.b": 10},
        {"NA.a": 3, "NA.b": 20},
        {"NA.a": 5, "NA.b": 10},
        {"NA.a": 5, "NA.b": 20},
    ]

    st2.prepare_inputs()
    assert st2.inputs_ind == [{"NB.a": 0}, {"NB.a": 1}, {"NB.a": 2}, {"NB.a": 3}]


def test_state_connect_6():
    """ three 'connected' states: testing groups, prepare_states and prepare_inputs,
        the first state has outer splitter,
        the third state has explicit splitter with splitters from previous states
     """
    st1 = State(name="NA", splitter=["a", "b"])
    st2 = State(name="NB", splitter="a")
    st3 = State(
        name="NC",
        splitter=["_NA", "_NB"],
        other_states={"NA": (st1, "a"), "NB": (st2, "b")},
    )
    assert st3.splitter == ["_NA", "_NB"]
    assert st3.splitter_rpn == ["NA.a", "NA.b", "*", "NB.a", "*"]

    st3.prepare_states(inputs={"NA.a": [3, 5], "NA.b": [10, 20], "NB.a": [600, 700]})
    assert st3.group_for_inputs_final == {"NA.a": 0, "NA.b": 1, "NB.a": 2}
    assert st3.groups_stack_final == [[0, 1, 2]]

    assert st3.states_ind == [
        {"NA.a": 0, "NA.b": 0, "NB.a": 0},
        {"NA.a": 0, "NA.b": 0, "NB.a": 1},
        {"NA.a": 0, "NA.b": 1, "NB.a": 0},
        {"NA.a": 0, "NA.b": 1, "NB.a": 1},
        {"NA.a": 1, "NA.b": 0, "NB.a": 0},
        {"NA.a": 1, "NA.b": 0, "NB.a": 1},
        {"NA.a": 1, "NA.b": 1, "NB.a": 0},
        {"NA.a": 1, "NA.b": 1, "NB.a": 1},
    ]
    assert st3.states_val == [
        {"NA.a": 3, "NA.b": 10, "NB.a": 600},
        {"NA.a": 3, "NA.b": 10, "NB.a": 700},
        {"NA.a": 3, "NA.b": 20, "NB.a": 600},
        {"NA.a": 3, "NA.b": 20, "NB.a": 700},
        {"NA.a": 5, "NA.b": 10, "NB.a": 600},
        {"NA.a": 5, "NA.b": 10, "NB.a": 700},
        {"NA.a": 5, "NA.b": 20, "NB.a": 600},
        {"NA.a": 5, "NA.b": 20, "NB.a": 700},
    ]

    st3.prepare_inputs()
    assert st3.inputs_ind == [
        {"NC.a": 0, "NC.b": 0},
        {"NC.a": 0, "NC.b": 1},
        {"NC.a": 1, "NC.b": 0},
        {"NC.a": 1, "NC.b": 1},
        {"NC.a": 2, "NC.b": 0},
        {"NC.a": 2, "NC.b": 1},
        {"NC.a": 3, "NC.b": 0},
        {"NC.a": 3, "NC.b": 1},
    ]


def test_state_connect_6a():
    """ three 'connected' states: testing groups, prepare_states and prepare_inputs,
        the first state has outer splitter,
        the third state has no explicit splitter
    """
    st1 = State(name="NA", splitter=["a", "b"])
    st2 = State(name="NB", splitter="a")
    st3 = State(name="NC", other_states={"NA": (st1, "a"), "NB": (st2, "b")})
    assert st3.splitter == ["_NA", "_NB"]
    assert st3.splitter_rpn == ["NA.a", "NA.b", "*", "NB.a", "*"]

    st3.prepare_states(inputs={"NA.a": [3, 5], "NA.b": [10, 20], "NB.a": [600, 700]})
    assert st3.group_for_inputs_final == {"NA.a": 0, "NA.b": 1, "NB.a": 2}
    assert st3.groups_stack_final == [[0, 1, 2]]
    assert st3.states_ind == [
        {"NA.a": 0, "NA.b": 0, "NB.a": 0},
        {"NA.a": 0, "NA.b": 0, "NB.a": 1},
        {"NA.a": 0, "NA.b": 1, "NB.a": 0},
        {"NA.a": 0, "NA.b": 1, "NB.a": 1},
        {"NA.a": 1, "NA.b": 0, "NB.a": 0},
        {"NA.a": 1, "NA.b": 0, "NB.a": 1},
        {"NA.a": 1, "NA.b": 1, "NB.a": 0},
        {"NA.a": 1, "NA.b": 1, "NB.a": 1},
    ]
    assert st3.states_val == [
        {"NA.a": 3, "NA.b": 10, "NB.a": 600},
        {"NA.a": 3, "NA.b": 10, "NB.a": 700},
        {"NA.a": 3, "NA.b": 20, "NB.a": 600},
        {"NA.a": 3, "NA.b": 20, "NB.a": 700},
        {"NA.a": 5, "NA.b": 10, "NB.a": 600},
        {"NA.a": 5, "NA.b": 10, "NB.a": 700},
        {"NA.a": 5, "NA.b": 20, "NB.a": 600},
        {"NA.a": 5, "NA.b": 20, "NB.a": 700},
    ]

    st3.prepare_inputs()
    assert st3.inputs_ind == [
        {"NC.a": 0, "NC.b": 0},
        {"NC.a": 0, "NC.b": 1},
        {"NC.a": 1, "NC.b": 0},
        {"NC.a": 1, "NC.b": 1},
        {"NC.a": 2, "NC.b": 0},
        {"NC.a": 2, "NC.b": 1},
        {"NC.a": 3, "NC.b": 0},
        {"NC.a": 3, "NC.b": 1},
    ]


def test_state_connect_innerspl_1():
    """ two 'connected' states: testing groups, prepare_states and prepare_inputs,
        the second state has an inner splitter, full splitter provided
    """
    st1 = State(name="NA", splitter="a")
    st2 = State(name="NB", splitter=["_NA", "b"], other_states={"NA": (st1, "b")})

    assert st2.splitter == ["_NA", "NB.b"]
    assert st2.splitter_rpn == ["NA.a", "NB.b", "*"]
    assert st2.left_splitter == "_NA"
    assert st2.left_splitter_rpn == ["NA.a"]
    assert st2.left_splitter_rpn_compact == ["_NA"]
    assert st2.right_splitter == "NB.b"
    assert st2.right_splitter_rpn == ["NB.b"]

    st2.prepare_states(
        inputs={"NA.a": [3, 5], "NB.b": [[1, 10, 100], [2, 20, 200]]},
        cont_dim={"NB.b": 2},  # will be treated as 2d container
    )
    assert st2.other_states["NA"][1] == "b"
    assert st2.group_for_inputs_final == {"NA.a": 0, "NB.b": 1}
    assert st2.groups_stack_final == [[0], [1]]

    assert st2.states_ind == [
        {"NA.a": 0, "NB.b": 0},
        {"NA.a": 0, "NB.b": 1},
        {"NA.a": 0, "NB.b": 2},
        {"NA.a": 1, "NB.b": 3},
        {"NA.a": 1, "NB.b": 4},
        {"NA.a": 1, "NB.b": 5},
    ]
    assert st2.states_val == [
        {"NA.a": 3, "NB.b": 1},
        {"NA.a": 3, "NB.b": 10},
        {"NA.a": 3, "NB.b": 100},
        {"NA.a": 5, "NB.b": 2},
        {"NA.a": 5, "NB.b": 20},
        {"NA.a": 5, "NB.b": 200},
    ]

    st2.prepare_inputs()
    assert st2.inputs_ind == [
        {"NB.b": 0},
        {"NB.b": 1},
        {"NB.b": 2},
        {"NB.b": 3},
        {"NB.b": 4},
        {"NB.b": 5},
    ]


def test_state_connect_innerspl_1a():
    """ two 'connected' states: testing groups, prepare_states and prepare_inputs,
        the second state has an inner splitter,
        splitter from the first state (Left part) has to be added
    """
    st1 = State(name="NA", splitter="a")
    st2 = State(name="NB", splitter="b", other_states={"NA": (st1, "b")})

    assert st2.splitter == ["_NA", "NB.b"]
    assert st2.splitter_rpn == ["NA.a", "NB.b", "*"]
    assert st2.left_splitter == "_NA"
    assert st2.left_splitter_rpn == ["NA.a"]
    assert st2.left_splitter_rpn_compact == ["_NA"]
    assert st2.right_splitter == "NB.b"
    assert st2.right_splitter_rpn == ["NB.b"]

    assert st2.other_states["NA"][1] == "b"

    st2.prepare_states(
        inputs={"NA.a": [3, 5], "NB.b": [[1, 10, 100], [2, 20, 200]]},
        cont_dim={"NB.b": 2},  # will be treated as 2d container
    )
    assert st2.group_for_inputs_final == {"NA.a": 0, "NB.b": 1}
    assert st2.groups_stack_final == [[0], [1]]

    assert st2.states_ind == [
        {"NA.a": 0, "NB.b": 0},
        {"NA.a": 0, "NB.b": 1},
        {"NA.a": 0, "NB.b": 2},
        {"NA.a": 1, "NB.b": 3},
        {"NA.a": 1, "NB.b": 4},
        {"NA.a": 1, "NB.b": 5},
    ]
    assert st2.states_val == [
        {"NA.a": 3, "NB.b": 1},
        {"NA.a": 3, "NB.b": 10},
        {"NA.a": 3, "NB.b": 100},
        {"NA.a": 5, "NB.b": 2},
        {"NA.a": 5, "NB.b": 20},
        {"NA.a": 5, "NB.b": 200},
    ]

    st2.prepare_inputs()
    assert st2.inputs_ind == [
        {"NB.b": 0},
        {"NB.b": 1},
        {"NB.b": 2},
        {"NB.b": 3},
        {"NB.b": 4},
        {"NB.b": 5},
    ]


def test_state_connect_innerspl_1b():
    """incorrect splitter - Right & Left parts in scalar splitter"""
    with pytest.raises(PydraStateError):
        st1 = State(name="NA", splitter="a")
        st2 = State(name="NB", splitter=("_NA", "b"), other_states={"NA": (st1, "b")})


def test_state_connect_innerspl_2():
    """ two 'connected' states: testing groups, prepare_states and prepare_inputs,
        the second state has one inner splitter and one 'normal' splitter
        only Right part of the splitter provided (Left has to be added)
    """
    st1 = State(name="NA", splitter="a")
    st2 = State(name="NB", splitter=["c", "b"], other_states={"NA": (st1, "b")})

    assert st2.splitter == ["_NA", ["NB.c", "NB.b"]]
    assert st2.splitter_rpn == ["NA.a", "NB.c", "NB.b", "*", "*"]
    assert st2.left_splitter == "_NA"
    assert st2.left_splitter_rpn == ["NA.a"]
    assert st2.left_splitter_rpn_compact == ["_NA"]
    assert st2.right_splitter == ["NB.c", "NB.b"]
    assert st2.right_splitter_rpn == ["NB.c", "NB.b", "*"]

    st2.prepare_states(
        inputs={"NA.a": [3, 5], "NB.b": [[1, 10, 100], [2, 20, 200]], "NB.c": [13, 17]},
        cont_dim={"NB.b": 2},  # will be treated as 2d container
    )
    assert st2.other_states["NA"][1] == "b"
    assert st2.group_for_inputs_final == {"NA.a": 0, "NB.c": 1, "NB.b": 2}
    assert st2.groups_stack_final == [[0], [1, 2]]

    assert st2.states_ind == [
        {"NB.c": 0, "NA.a": 0, "NB.b": 0},
        {"NB.c": 0, "NA.a": 0, "NB.b": 1},
        {"NB.c": 0, "NA.a": 0, "NB.b": 2},
        {"NB.c": 0, "NA.a": 1, "NB.b": 3},
        {"NB.c": 0, "NA.a": 1, "NB.b": 4},
        {"NB.c": 0, "NA.a": 1, "NB.b": 5},
        {"NB.c": 1, "NA.a": 0, "NB.b": 0},
        {"NB.c": 1, "NA.a": 0, "NB.b": 1},
        {"NB.c": 1, "NA.a": 0, "NB.b": 2},
        {"NB.c": 1, "NA.a": 1, "NB.b": 3},
        {"NB.c": 1, "NA.a": 1, "NB.b": 4},
        {"NB.c": 1, "NA.a": 1, "NB.b": 5},
    ]
    assert st2.states_val == [
        {"NB.c": 13, "NA.a": 3, "NB.b": 1},
        {"NB.c": 13, "NA.a": 3, "NB.b": 10},
        {"NB.c": 13, "NA.a": 3, "NB.b": 100},
        {"NB.c": 13, "NA.a": 5, "NB.b": 2},
        {"NB.c": 13, "NA.a": 5, "NB.b": 20},
        {"NB.c": 13, "NA.a": 5, "NB.b": 200},
        {"NB.c": 17, "NA.a": 3, "NB.b": 1},
        {"NB.c": 17, "NA.a": 3, "NB.b": 10},
        {"NB.c": 17, "NA.a": 3, "NB.b": 100},
        {"NB.c": 17, "NA.a": 5, "NB.b": 2},
        {"NB.c": 17, "NA.a": 5, "NB.b": 20},
        {"NB.c": 17, "NA.a": 5, "NB.b": 200},
    ]

    st2.prepare_inputs()
    assert st2.inputs_ind == [
        {"NB.c": 0, "NB.b": 0},
        {"NB.c": 0, "NB.b": 1},
        {"NB.c": 0, "NB.b": 2},
        {"NB.c": 0, "NB.b": 3},
        {"NB.c": 0, "NB.b": 4},
        {"NB.c": 0, "NB.b": 5},
        {"NB.c": 1, "NB.b": 0},
        {"NB.c": 1, "NB.b": 1},
        {"NB.c": 1, "NB.b": 2},
        {"NB.c": 1, "NB.b": 3},
        {"NB.c": 1, "NB.b": 4},
        {"NB.c": 1, "NB.b": 5},
    ]


def test_state_connect_innerspl_2a():
    """ two 'connected' states: testing groups, prepare_states and prepare_inputs,
        the second state has one inner splitter and one 'normal' splitter
        only Right part of the splitter provided (different order!),

    """
    st1 = State(name="NA", splitter="a")
    st2 = State(name="NB", splitter=["b", "c"], other_states={"NA": (st1, "b")})

    assert st2.splitter == ["_NA", ["NB.b", "NB.c"]]
    assert st2.splitter_rpn == ["NA.a", "NB.b", "NB.c", "*", "*"]
    assert st2.other_states["NA"][1] == "b"

    st2.prepare_states(
        inputs={"NA.a": [3, 5], "NB.b": [[1, 10, 100], [2, 20, 200]], "NB.c": [13, 17]},
        cont_dim={"NB.b": 2},  # will be treated as 2d container
    )
    assert st2.group_for_inputs_final == {"NA.a": 0, "NB.c": 2, "NB.b": 1}
    assert st2.groups_stack_final == [[0], [1, 2]]

    assert st2.states_ind == [
        {"NB.c": 0, "NA.a": 0, "NB.b": 0},
        {"NB.c": 1, "NA.a": 0, "NB.b": 0},
        {"NB.c": 0, "NA.a": 0, "NB.b": 1},
        {"NB.c": 1, "NA.a": 0, "NB.b": 1},
        {"NB.c": 0, "NA.a": 0, "NB.b": 2},
        {"NB.c": 1, "NA.a": 0, "NB.b": 2},
        {"NB.c": 0, "NA.a": 1, "NB.b": 3},
        {"NB.c": 1, "NA.a": 1, "NB.b": 3},
        {"NB.c": 0, "NA.a": 1, "NB.b": 4},
        {"NB.c": 1, "NA.a": 1, "NB.b": 4},
        {"NB.c": 0, "NA.a": 1, "NB.b": 5},
        {"NB.c": 1, "NA.a": 1, "NB.b": 5},
    ]

    assert st2.states_val == [
        {"NB.c": 13, "NA.a": 3, "NB.b": 1},
        {"NB.c": 17, "NA.a": 3, "NB.b": 1},
        {"NB.c": 13, "NA.a": 3, "NB.b": 10},
        {"NB.c": 17, "NA.a": 3, "NB.b": 10},
        {"NB.c": 13, "NA.a": 3, "NB.b": 100},
        {"NB.c": 17, "NA.a": 3, "NB.b": 100},
        {"NB.c": 13, "NA.a": 5, "NB.b": 2},
        {"NB.c": 17, "NA.a": 5, "NB.b": 2},
        {"NB.c": 13, "NA.a": 5, "NB.b": 20},
        {"NB.c": 17, "NA.a": 5, "NB.b": 20},
        {"NB.c": 13, "NA.a": 5, "NB.b": 200},
        {"NB.c": 17, "NA.a": 5, "NB.b": 200},
    ]

    st2.prepare_inputs()
    assert st2.inputs_ind == [
        {"NB.c": 0, "NB.b": 0},
        {"NB.c": 1, "NB.b": 0},
        {"NB.c": 0, "NB.b": 1},
        {"NB.c": 1, "NB.b": 1},
        {"NB.c": 0, "NB.b": 2},
        {"NB.c": 1, "NB.b": 2},
        {"NB.c": 0, "NB.b": 3},
        {"NB.c": 1, "NB.b": 3},
        {"NB.c": 0, "NB.b": 4},
        {"NB.c": 1, "NB.b": 4},
        {"NB.c": 0, "NB.b": 5},
        {"NB.c": 1, "NB.b": 5},
    ]


def test_state_connect_innerspl_3():
    """ three serially 'connected' states: testing groups, prepare_states and prepare_inputs,
        the second state has one inner splitter and one 'normal' splitter
        Left parts have to be added
    """
    st1 = State(name="NA", splitter="a")
    st2 = State(name="NB", splitter=["c", "b"], other_states={"NA": (st1, "b")})
    st3 = State(name="NC", splitter="d", other_states={"NB": (st2, "a")})

    assert st3.splitter == ["_NB", "NC.d"]
    assert st3.splitter_rpn == ["NA.a", "NB.c", "NB.b", "*", "*", "NC.d", "*"]
    assert st3.left_splitter == "_NB"
    assert st3.left_splitter_rpn == ["NA.a", "NB.c", "NB.b", "*", "*"]
    assert st3.left_splitter_rpn_compact == ["_NB"]
    assert st3.right_splitter == "NC.d"
    assert st3.right_splitter_rpn == ["NC.d"]

    st3.prepare_states(
        inputs={
            "NA.a": [3, 5],
            "NB.b": [[1, 10, 100], [2, 20, 200]],
            "NB.c": [13, 17],
            "NC.d": [33, 77],
        },
        cont_dim={"NB.b": 2},  # will be treated as 2d container
    )
    assert st3.group_for_inputs_final == {"NA.a": 0, "NB.c": 1, "NB.b": 2, "NC.d": 3}
    assert st3.groups_stack_final == [[0], [1, 2, 3]]

    assert st2.states_ind == [
        {"NB.c": 0, "NA.a": 0, "NB.b": 0},
        {"NB.c": 0, "NA.a": 0, "NB.b": 1},
        {"NB.c": 0, "NA.a": 0, "NB.b": 2},
        {"NB.c": 0, "NA.a": 1, "NB.b": 3},
        {"NB.c": 0, "NA.a": 1, "NB.b": 4},
        {"NB.c": 0, "NA.a": 1, "NB.b": 5},
        {"NB.c": 1, "NA.a": 0, "NB.b": 0},
        {"NB.c": 1, "NA.a": 0, "NB.b": 1},
        {"NB.c": 1, "NA.a": 0, "NB.b": 2},
        {"NB.c": 1, "NA.a": 1, "NB.b": 3},
        {"NB.c": 1, "NA.a": 1, "NB.b": 4},
        {"NB.c": 1, "NA.a": 1, "NB.b": 5},
    ]
    assert st2.states_val == [
        {"NB.c": 13, "NA.a": 3, "NB.b": 1},
        {"NB.c": 13, "NA.a": 3, "NB.b": 10},
        {"NB.c": 13, "NA.a": 3, "NB.b": 100},
        {"NB.c": 13, "NA.a": 5, "NB.b": 2},
        {"NB.c": 13, "NA.a": 5, "NB.b": 20},
        {"NB.c": 13, "NA.a": 5, "NB.b": 200},
        {"NB.c": 17, "NA.a": 3, "NB.b": 1},
        {"NB.c": 17, "NA.a": 3, "NB.b": 10},
        {"NB.c": 17, "NA.a": 3, "NB.b": 100},
        {"NB.c": 17, "NA.a": 5, "NB.b": 2},
        {"NB.c": 17, "NA.a": 5, "NB.b": 20},
        {"NB.c": 17, "NA.a": 5, "NB.b": 200},
    ]

    assert st3.states_ind == [
        {"NB.c": 0, "NA.a": 0, "NB.b": 0, "NC.d": 0},
        {"NB.c": 0, "NA.a": 0, "NB.b": 0, "NC.d": 1},
        {"NB.c": 0, "NA.a": 0, "NB.b": 1, "NC.d": 0},
        {"NB.c": 0, "NA.a": 0, "NB.b": 1, "NC.d": 1},
        {"NB.c": 0, "NA.a": 0, "NB.b": 2, "NC.d": 0},
        {"NB.c": 0, "NA.a": 0, "NB.b": 2, "NC.d": 1},
        {"NB.c": 0, "NA.a": 1, "NB.b": 3, "NC.d": 0},
        {"NB.c": 0, "NA.a": 1, "NB.b": 3, "NC.d": 1},
        {"NB.c": 0, "NA.a": 1, "NB.b": 4, "NC.d": 0},
        {"NB.c": 0, "NA.a": 1, "NB.b": 4, "NC.d": 1},
        {"NB.c": 0, "NA.a": 1, "NB.b": 5, "NC.d": 0},
        {"NB.c": 0, "NA.a": 1, "NB.b": 5, "NC.d": 1},
        {"NB.c": 1, "NA.a": 0, "NB.b": 0, "NC.d": 0},
        {"NB.c": 1, "NA.a": 0, "NB.b": 0, "NC.d": 1},
        {"NB.c": 1, "NA.a": 0, "NB.b": 1, "NC.d": 0},
        {"NB.c": 1, "NA.a": 0, "NB.b": 1, "NC.d": 1},
        {"NB.c": 1, "NA.a": 0, "NB.b": 2, "NC.d": 0},
        {"NB.c": 1, "NA.a": 0, "NB.b": 2, "NC.d": 1},
        {"NB.c": 1, "NA.a": 1, "NB.b": 3, "NC.d": 0},
        {"NB.c": 1, "NA.a": 1, "NB.b": 3, "NC.d": 1},
        {"NB.c": 1, "NA.a": 1, "NB.b": 4, "NC.d": 0},
        {"NB.c": 1, "NA.a": 1, "NB.b": 4, "NC.d": 1},
        {"NB.c": 1, "NA.a": 1, "NB.b": 5, "NC.d": 0},
        {"NB.c": 1, "NA.a": 1, "NB.b": 5, "NC.d": 1},
    ]
    assert st3.states_val == [
        {"NB.c": 13, "NA.a": 3, "NB.b": 1, "NC.d": 33},
        {"NB.c": 13, "NA.a": 3, "NB.b": 1, "NC.d": 77},
        {"NB.c": 13, "NA.a": 3, "NB.b": 10, "NC.d": 33},
        {"NB.c": 13, "NA.a": 3, "NB.b": 10, "NC.d": 77},
        {"NB.c": 13, "NA.a": 3, "NB.b": 100, "NC.d": 33},
        {"NB.c": 13, "NA.a": 3, "NB.b": 100, "NC.d": 77},
        {"NB.c": 13, "NA.a": 5, "NB.b": 2, "NC.d": 33},
        {"NB.c": 13, "NA.a": 5, "NB.b": 2, "NC.d": 77},
        {"NB.c": 13, "NA.a": 5, "NB.b": 20, "NC.d": 33},
        {"NB.c": 13, "NA.a": 5, "NB.b": 20, "NC.d": 77},
        {"NB.c": 13, "NA.a": 5, "NB.b": 200, "NC.d": 33},
        {"NB.c": 13, "NA.a": 5, "NB.b": 200, "NC.d": 77},
        {"NB.c": 17, "NA.a": 3, "NB.b": 1, "NC.d": 33},
        {"NB.c": 17, "NA.a": 3, "NB.b": 1, "NC.d": 77},
        {"NB.c": 17, "NA.a": 3, "NB.b": 10, "NC.d": 33},
        {"NB.c": 17, "NA.a": 3, "NB.b": 10, "NC.d": 77},
        {"NB.c": 17, "NA.a": 3, "NB.b": 100, "NC.d": 33},
        {"NB.c": 17, "NA.a": 3, "NB.b": 100, "NC.d": 77},
        {"NB.c": 17, "NA.a": 5, "NB.b": 2, "NC.d": 33},
        {"NB.c": 17, "NA.a": 5, "NB.b": 2, "NC.d": 77},
        {"NB.c": 17, "NA.a": 5, "NB.b": 20, "NC.d": 33},
        {"NB.c": 17, "NA.a": 5, "NB.b": 20, "NC.d": 77},
        {"NB.c": 17, "NA.a": 5, "NB.b": 200, "NC.d": 33},
        {"NB.c": 17, "NA.a": 5, "NB.b": 200, "NC.d": 77},
    ]

    st3.prepare_inputs()
    assert st3.inputs_ind == [
        {"NC.a": 0, "NC.d": 0},
        {"NC.a": 0, "NC.d": 1},
        {"NC.a": 1, "NC.d": 0},
        {"NC.a": 1, "NC.d": 1},
        {"NC.a": 2, "NC.d": 0},
        {"NC.a": 2, "NC.d": 1},
        {"NC.a": 3, "NC.d": 0},
        {"NC.a": 3, "NC.d": 1},
        {"NC.a": 4, "NC.d": 0},
        {"NC.a": 4, "NC.d": 1},
        {"NC.a": 5, "NC.d": 0},
        {"NC.a": 5, "NC.d": 1},
        {"NC.a": 6, "NC.d": 0},
        {"NC.a": 6, "NC.d": 1},
        {"NC.a": 7, "NC.d": 0},
        {"NC.a": 7, "NC.d": 1},
        {"NC.a": 8, "NC.d": 0},
        {"NC.a": 8, "NC.d": 1},
        {"NC.a": 9, "NC.d": 0},
        {"NC.a": 9, "NC.d": 1},
        {"NC.a": 10, "NC.d": 0},
        {"NC.a": 10, "NC.d": 1},
        {"NC.a": 11, "NC.d": 0},
        {"NC.a": 11, "NC.d": 1},
    ]


def test_state_connect_innerspl_4():
    """ three'connected' states: testing groups, prepare_states and prepare_inputs,
        the third one connected to two previous, only Right part of splitter provided
    """
    st1 = State(name="NA", splitter="a")
    st2 = State(name="NB", splitter=["b", "c"])
    st3 = State(
        name="NC", splitter="d", other_states={"NA": (st1, "e"), "NB": (st2, "f")}
    )

    assert st3.splitter == [["_NA", "_NB"], "NC.d"]
    assert st3.splitter_rpn == ["NA.a", "NB.b", "NB.c", "*", "*", "NC.d", "*"]
    assert st3.other_states["NA"][1] == "e"
    assert st3.other_states["NB"][1] == "f"

    st3.prepare_states(
        inputs={
            "NA.a": [3, 5],
            "NB.b": [10, 20],
            "NB.c": [13, 17],
            "NC.e": [30, 50],
            "NC.f": [[23, 27], [33, 37]],
            "NC.d": [1, 2],
        },
        cont_dim={"NC.f": 2},  # will be treated as 2d container
    )
    assert st3.group_for_inputs_final == {"NA.a": 0, "NB.c": 2, "NB.b": 1, "NC.d": 3}
    assert st3.groups_stack_final == [[0, 1, 2, 3]]
    assert st3.states_ind == [
        {"NA.a": 0, "NB.b": 0, "NB.c": 0, "NC.d": 0},
        {"NA.a": 0, "NB.b": 0, "NB.c": 0, "NC.d": 1},
        {"NA.a": 0, "NB.b": 0, "NB.c": 1, "NC.d": 0},
        {"NA.a": 0, "NB.b": 0, "NB.c": 1, "NC.d": 1},
        {"NA.a": 0, "NB.b": 1, "NB.c": 0, "NC.d": 0},
        {"NA.a": 0, "NB.b": 1, "NB.c": 0, "NC.d": 1},
        {"NA.a": 0, "NB.b": 1, "NB.c": 1, "NC.d": 0},
        {"NA.a": 0, "NB.b": 1, "NB.c": 1, "NC.d": 1},
        {"NA.a": 1, "NB.b": 0, "NB.c": 0, "NC.d": 0},
        {"NA.a": 1, "NB.b": 0, "NB.c": 0, "NC.d": 1},
        {"NA.a": 1, "NB.b": 0, "NB.c": 1, "NC.d": 0},
        {"NA.a": 1, "NB.b": 0, "NB.c": 1, "NC.d": 1},
        {"NA.a": 1, "NB.b": 1, "NB.c": 0, "NC.d": 0},
        {"NA.a": 1, "NB.b": 1, "NB.c": 0, "NC.d": 1},
        {"NA.a": 1, "NB.b": 1, "NB.c": 1, "NC.d": 0},
        {"NA.a": 1, "NB.b": 1, "NB.c": 1, "NC.d": 1},
    ]

    assert st3.states_val == [
        {"NA.a": 3, "NB.b": 10, "NB.c": 13, "NC.d": 1},
        {"NA.a": 3, "NB.b": 10, "NB.c": 13, "NC.d": 2},
        {"NA.a": 3, "NB.b": 10, "NB.c": 17, "NC.d": 1},
        {"NA.a": 3, "NB.b": 10, "NB.c": 17, "NC.d": 2},
        {"NA.a": 3, "NB.b": 20, "NB.c": 13, "NC.d": 1},
        {"NA.a": 3, "NB.b": 20, "NB.c": 13, "NC.d": 2},
        {"NA.a": 3, "NB.b": 20, "NB.c": 17, "NC.d": 1},
        {"NA.a": 3, "NB.b": 20, "NB.c": 17, "NC.d": 2},
        {"NA.a": 5, "NB.b": 10, "NB.c": 13, "NC.d": 1},
        {"NA.a": 5, "NB.b": 10, "NB.c": 13, "NC.d": 2},
        {"NA.a": 5, "NB.b": 10, "NB.c": 17, "NC.d": 1},
        {"NA.a": 5, "NB.b": 10, "NB.c": 17, "NC.d": 2},
        {"NA.a": 5, "NB.b": 20, "NB.c": 13, "NC.d": 1},
        {"NA.a": 5, "NB.b": 20, "NB.c": 13, "NC.d": 2},
        {"NA.a": 5, "NB.b": 20, "NB.c": 17, "NC.d": 1},
        {"NA.a": 5, "NB.b": 20, "NB.c": 17, "NC.d": 2},
    ]

    st3.prepare_inputs()
    assert st3.inputs_ind == [
        {"NC.e": 0, "NC.f": 0, "NC.d": 0},
        {"NC.e": 0, "NC.f": 0, "NC.d": 1},
        {"NC.e": 0, "NC.f": 1, "NC.d": 0},
        {"NC.e": 0, "NC.f": 1, "NC.d": 1},
        {"NC.e": 0, "NC.f": 2, "NC.d": 0},
        {"NC.e": 0, "NC.f": 2, "NC.d": 1},
        {"NC.e": 0, "NC.f": 3, "NC.d": 0},
        {"NC.e": 0, "NC.f": 3, "NC.d": 1},
        {"NC.e": 1, "NC.f": 0, "NC.d": 0},
        {"NC.e": 1, "NC.f": 0, "NC.d": 1},
        {"NC.e": 1, "NC.f": 1, "NC.d": 0},
        {"NC.e": 1, "NC.f": 1, "NC.d": 1},
        {"NC.e": 1, "NC.f": 2, "NC.d": 0},
        {"NC.e": 1, "NC.f": 2, "NC.d": 1},
        {"NC.e": 1, "NC.f": 3, "NC.d": 0},
        {"NC.e": 1, "NC.f": 3, "NC.d": 1},
    ]


def test_state_combine_1():
    """ single state with splitter and combiner"""
    st = State(name="NA", splitter="a", combiner="a")
    assert st.splitter == "NA.a"
    assert st.splitter_rpn == ["NA.a"]
    assert st.right_combiner == st.right_combiner_all == st.combiner == ["NA.a"]
    assert st.left_combiner == st.left_combiner_all == []
    assert st.splitter_final == None
    assert st.splitter_rpn_final == []

    st.prepare_states(inputs={"NA.a": [3, 5]})
    assert st.group_for_inputs_final == {}
    assert st.groups_stack_final == []
    assert st.states_ind == [{"NA.a": 0}, {"NA.a": 1}]
    assert st.states_val == [{"NA.a": 3}, {"NA.a": 5}]

    st.prepare_inputs()
    assert st.inputs_ind == [{"NA.a": 0}, {"NA.a": 1}]


def test_state_connect_combine_1():
    """two connected states; outer splitter and combiner in the first one"""
    st1 = State(name="NA", splitter=["a", "b"], combiner="a")
    st2 = State(name="NB", other_states={"NA": (st1, "c")})

    assert st1.splitter == ["NA.a", "NA.b"]
    assert st1.splitter_rpn == ["NA.a", "NA.b", "*"]
    assert st1.splitter_rpn_final == ["NA.b"]
    assert st1.splitter_final == "NA.b"
    assert st1.combiner == st1.right_combiner == st1.right_combiner_all == ["NA.a"]
    assert st1.left_combiner_all == st1.left_combiner == []
    assert st2.splitter == "_NA"
    assert st2.splitter_rpn == ["NA.b"]

    st2.prepare_states(inputs={"NA.a": [3, 5], "NA.b": [10, 20]})
    assert st2.group_for_inputs_final == {"NA.b": 0}
    assert st2.groups_stack_final == [[0]]
    assert st1.states_ind == [
        {"NA.a": 0, "NA.b": 0},
        {"NA.a": 0, "NA.b": 1},
        {"NA.a": 1, "NA.b": 0},
        {"NA.a": 1, "NA.b": 1},
    ]
    assert st1.states_val == [
        {"NA.a": 3, "NA.b": 10},
        {"NA.a": 3, "NA.b": 20},
        {"NA.a": 5, "NA.b": 10},
        {"NA.a": 5, "NA.b": 20},
    ]
    assert st1.states_ind_final == [{"NA.b": 0}, {"NA.b": 1}]
    assert st1.keys_final == ["NA.b"]
    assert st1.final_combined_ind_mapping == {0: [0, 2], 1: [1, 3]}

    assert st2.states_ind == [{"NA.b": 0}, {"NA.b": 1}]
    assert st2.states_val == [{"NA.b": 10}, {"NA.b": 20}]
    assert st2.keys_final == ["NA.b"]
    assert st2.final_combined_ind_mapping == {0: [0], 1: [1]}

    st2.prepare_inputs()
    assert st2.inputs_ind == [{"NB.c": 0}, {"NB.c": 1}]


def test_state_connect_combine_2():
    """
    two connected states; outer splitter and combiner in the first one;
    additional splitter in the second node
    """
    st1 = State(name="NA", splitter=["a", "b"], combiner="a")
    st2 = State(name="NB", splitter="d", other_states={"NA": (st1, "c")})

    assert st1.splitter == ["NA.a", "NA.b"]
    assert st1.splitter_rpn == ["NA.a", "NA.b", "*"]
    assert st1.combiner == ["NA.a"]
    assert st1.splitter_final == "NA.b"
    assert st1.splitter_rpn_final == ["NA.b"]

    assert st2.splitter == ["_NA", "NB.d"]
    assert st2.splitter_rpn == ["NA.b", "NB.d", "*"]

    st2.prepare_states(
        inputs={"NA.a": [3, 5], "NA.b": [10, 20], "NB.c": [90, 150], "NB.d": [0, 1]}
    )
    assert st2.group_for_inputs_final == {"NA.b": 0, "NB.d": 1}
    assert st2.groups_stack_final == [[0, 1]]
    assert st1.states_ind == [
        {"NA.a": 0, "NA.b": 0},
        {"NA.a": 0, "NA.b": 1},
        {"NA.a": 1, "NA.b": 0},
        {"NA.a": 1, "NA.b": 1},
    ]
    assert st1.states_val == [
        {"NA.a": 3, "NA.b": 10},
        {"NA.a": 3, "NA.b": 20},
        {"NA.a": 5, "NA.b": 10},
        {"NA.a": 5, "NA.b": 20},
    ]
    assert st1.states_ind_final == [{"NA.b": 0}, {"NA.b": 1}]
    assert st1.keys_final == ["NA.b"]
    assert st1.final_combined_ind_mapping == {0: [0, 2], 1: [1, 3]}

    assert st2.states_ind == [
        {"NA.b": 0, "NB.d": 0},
        {"NA.b": 0, "NB.d": 1},
        {"NA.b": 1, "NB.d": 0},
        {"NA.b": 1, "NB.d": 1},
    ]
    assert st2.states_val == [
        {"NA.b": 10, "NB.d": 0},
        {"NA.b": 10, "NB.d": 1},
        {"NA.b": 20, "NB.d": 0},
        {"NA.b": 20, "NB.d": 1},
    ]
    assert st2.keys_final == ["NA.b", "NB.d"]
    assert st2.final_combined_ind_mapping == {0: [0], 1: [1], 2: [2], 3: [3]}

    st2.prepare_inputs()
    assert st2.inputs_ind == [
        {"NB.c": 0, "NB.d": 0},
        {"NB.c": 0, "NB.d": 1},
        {"NB.c": 1, "NB.d": 0},
        {"NB.c": 1, "NB.d": 1},
    ]


def test_state_connect_combine_3():
    """
    two connected states; outer splitter and combiner in the first one;
    additional splitter in the second node
    """
    st1 = State(name="NA", splitter=["a", "b"], combiner="a")
    st2 = State(name="NB", splitter="d", combiner="d", other_states={"NA": (st1, "c")})

    assert st1.splitter == ["NA.a", "NA.b"]
    assert st1.splitter_rpn == ["NA.a", "NA.b", "*"]
    assert st1.combiner == ["NA.a"]
    assert st1.splitter_final == "NA.b"
    assert st1.splitter_rpn_final == ["NA.b"]

    assert st2.splitter == ["_NA", "NB.d"]
    assert st2.splitter_rpn == ["NA.b", "NB.d", "*"]
    assert st2.splitter_rpn_final == ["NA.b"]
    assert st2.left_combiner_all == st2.left_combiner == []
    assert st2.right_combiner_all == st2.right_combiner == st2.combiner == ["NB.d"]

    st2.prepare_states(
        inputs={"NA.a": [3, 5], "NA.b": [10, 20], "NB.c": [90, 150], "NB.d": [0, 1]}
    )
    assert st2.group_for_inputs_final == {"NA.b": 0}
    assert st2.groups_stack_final == [[0]]

    assert st1.states_ind == [
        {"NA.a": 0, "NA.b": 0},
        {"NA.a": 0, "NA.b": 1},
        {"NA.a": 1, "NA.b": 0},
        {"NA.a": 1, "NA.b": 1},
    ]
    assert st1.states_val == [
        {"NA.a": 3, "NA.b": 10},
        {"NA.a": 3, "NA.b": 20},
        {"NA.a": 5, "NA.b": 10},
        {"NA.a": 5, "NA.b": 20},
    ]
    assert st1.states_ind_final == [{"NA.b": 0}, {"NA.b": 1}]
    assert st1.keys_final == ["NA.b"]
    assert st1.final_combined_ind_mapping == {0: [0, 2], 1: [1, 3]}

    assert st2.states_ind == [
        {"NA.b": 0, "NB.d": 0},
        {"NA.b": 0, "NB.d": 1},
        {"NA.b": 1, "NB.d": 0},
        {"NA.b": 1, "NB.d": 1},
    ]
    assert st2.states_val == [
        {"NA.b": 10, "NB.d": 0},
        {"NA.b": 10, "NB.d": 1},
        {"NA.b": 20, "NB.d": 0},
        {"NA.b": 20, "NB.d": 1},
    ]
    assert st2.states_ind_final == [{"NA.b": 0}, {"NA.b": 1}]
    assert st2.keys_final == ["NA.b"]
    assert st2.final_combined_ind_mapping == {0: [0, 1], 1: [2, 3]}

    st2.prepare_inputs()
    assert st2.inputs_ind == [
        {"NB.c": 0, "NB.d": 0},
        {"NB.c": 0, "NB.d": 1},
        {"NB.c": 1, "NB.d": 0},
        {"NB.c": 1, "NB.d": 1},
    ]


def test_state_connect_innerspl_combine_1():
    """one previous node and one inner splitter (and inner splitter combiner);
    only Right part provided - Left had to be added"""
    st1 = State(name="NA", splitter="a")
    st2 = State(
        name="NB", splitter=["c", "b"], combiner=["b"], other_states={"NA": (st1, "b")}
    )

    assert st2.splitter == ["_NA", ["NB.c", "NB.b"]]
    assert st2.splitter_rpn == ["NA.a", "NB.c", "NB.b", "*", "*"]
    assert st2.splitter_final == ["NA.a", "NB.c"]
    assert st2.splitter_rpn_final == ["NA.a", "NB.c", "*"]
    assert st2.left_combiner_all == st2.left_combiner == []
    assert st2.right_combiner_all == st2.right_combiner == st2.combiner == ["NB.b"]
    # TODO: i think at the end I should merge [0] and [1], because there are no inner splitters anymore
    # TODO: didn't include it in my code...
    # assert st2.groups_stack_final == [[0, 1]]

    st2.prepare_states(
        inputs={"NA.a": [3, 5], "NB.b": [[1, 10, 100], [2, 20, 200]], "NB.c": [13, 17]},
        cont_dim={"NB.b": 2},  # will be treated as 2d container
    )
    assert st2.group_for_inputs_final == {"NA.a": 0, "NB.c": 1}
    assert st2.groups_stack_final == [[0], [1]]

    # NOW TODO: checking st2.states_ind_final!!!
    assert st2.states_ind == [
        {"NB.c": 0, "NA.a": 0, "NB.b": 0},
        {"NB.c": 0, "NA.a": 0, "NB.b": 1},
        {"NB.c": 0, "NA.a": 0, "NB.b": 2},
        {"NB.c": 0, "NA.a": 1, "NB.b": 3},
        {"NB.c": 0, "NA.a": 1, "NB.b": 4},
        {"NB.c": 0, "NA.a": 1, "NB.b": 5},
        {"NB.c": 1, "NA.a": 0, "NB.b": 0},
        {"NB.c": 1, "NA.a": 0, "NB.b": 1},
        {"NB.c": 1, "NA.a": 0, "NB.b": 2},
        {"NB.c": 1, "NA.a": 1, "NB.b": 3},
        {"NB.c": 1, "NA.a": 1, "NB.b": 4},
        {"NB.c": 1, "NA.a": 1, "NB.b": 5},
    ]
    assert st2.states_val == [
        {"NB.c": 13, "NA.a": 3, "NB.b": 1},
        {"NB.c": 13, "NA.a": 3, "NB.b": 10},
        {"NB.c": 13, "NA.a": 3, "NB.b": 100},
        {"NB.c": 13, "NA.a": 5, "NB.b": 2},
        {"NB.c": 13, "NA.a": 5, "NB.b": 20},
        {"NB.c": 13, "NA.a": 5, "NB.b": 200},
        {"NB.c": 17, "NA.a": 3, "NB.b": 1},
        {"NB.c": 17, "NA.a": 3, "NB.b": 10},
        {"NB.c": 17, "NA.a": 3, "NB.b": 100},
        {"NB.c": 17, "NA.a": 5, "NB.b": 2},
        {"NB.c": 17, "NA.a": 5, "NB.b": 20},
        {"NB.c": 17, "NA.a": 5, "NB.b": 200},
    ]
    assert st2.states_ind_final == [
        {"NB.c": 0, "NA.a": 0},
        {"NB.c": 1, "NA.a": 0},
        {"NB.c": 0, "NA.a": 1},
        {"NB.c": 1, "NA.a": 1},
    ]

    st2.prepare_inputs()
    assert st2.inputs_ind == [
        {"NB.c": 0, "NB.b": 0},
        {"NB.c": 0, "NB.b": 1},
        {"NB.c": 0, "NB.b": 2},
        {"NB.c": 0, "NB.b": 3},
        {"NB.c": 0, "NB.b": 4},
        {"NB.c": 0, "NB.b": 5},
        {"NB.c": 1, "NB.b": 0},
        {"NB.c": 1, "NB.b": 1},
        {"NB.c": 1, "NB.b": 2},
        {"NB.c": 1, "NB.b": 3},
        {"NB.c": 1, "NB.b": 4},
        {"NB.c": 1, "NB.b": 5},
    ]


def test_state_connect_innerspl_combine_2():
    """ two 'connected' state, the second has inner and normal splitter,
        and 'normal' combiner
        only Right part splitter provided, Left has to be added
        """
    st1 = State(name="NA", splitter="a")
    st2 = State(
        name="NB", splitter=["c", "b"], combiner=["c"], other_states={"NA": (st1, "b")}
    )

    assert st2.splitter == ["_NA", ["NB.c", "NB.b"]]
    assert st2.splitter_rpn == ["NA.a", "NB.c", "NB.b", "*", "*"]
    assert st2.splitter_final == ["NA.a", "NB.b"]
    assert st2.splitter_rpn_final == ["NA.a", "NB.b", "*"]

    st2.prepare_states(
        inputs={"NA.a": [3, 5], "NB.b": [[1, 10, 100], [2, 20, 200]], "NB.c": [13, 17]},
        cont_dim={"NB.b": 2},  # will be treated as 2d container
    )
    assert st2.group_for_inputs_final == {"NA.a": 0, "NB.b": 1}
    assert st2.groups_stack_final == [[0], [1]]
    assert st2.states_ind == [
        {"NB.c": 0, "NA.a": 0, "NB.b": 0},
        {"NB.c": 0, "NA.a": 0, "NB.b": 1},
        {"NB.c": 0, "NA.a": 0, "NB.b": 2},
        {"NB.c": 0, "NA.a": 1, "NB.b": 3},
        {"NB.c": 0, "NA.a": 1, "NB.b": 4},
        {"NB.c": 0, "NA.a": 1, "NB.b": 5},
        {"NB.c": 1, "NA.a": 0, "NB.b": 0},
        {"NB.c": 1, "NA.a": 0, "NB.b": 1},
        {"NB.c": 1, "NA.a": 0, "NB.b": 2},
        {"NB.c": 1, "NA.a": 1, "NB.b": 3},
        {"NB.c": 1, "NA.a": 1, "NB.b": 4},
        {"NB.c": 1, "NA.a": 1, "NB.b": 5},
    ]
    assert st2.states_val == [
        {"NB.c": 13, "NA.a": 3, "NB.b": 1},
        {"NB.c": 13, "NA.a": 3, "NB.b": 10},
        {"NB.c": 13, "NA.a": 3, "NB.b": 100},
        {"NB.c": 13, "NA.a": 5, "NB.b": 2},
        {"NB.c": 13, "NA.a": 5, "NB.b": 20},
        {"NB.c": 13, "NA.a": 5, "NB.b": 200},
        {"NB.c": 17, "NA.a": 3, "NB.b": 1},
        {"NB.c": 17, "NA.a": 3, "NB.b": 10},
        {"NB.c": 17, "NA.a": 3, "NB.b": 100},
        {"NB.c": 17, "NA.a": 5, "NB.b": 2},
        {"NB.c": 17, "NA.a": 5, "NB.b": 20},
        {"NB.c": 17, "NA.a": 5, "NB.b": 200},
    ]
    assert st2.states_ind_final == [
        {"NA.a": 0, "NB.b": 0},
        {"NA.a": 0, "NB.b": 1},
        {"NA.a": 0, "NB.b": 2},
        {"NA.a": 1, "NB.b": 3},
        {"NA.a": 1, "NB.b": 4},
        {"NA.a": 1, "NB.b": 5},
    ]

    st2.prepare_inputs()
    assert st2.inputs_ind == [
        {"NB.c": 0, "NB.b": 0},
        {"NB.c": 0, "NB.b": 1},
        {"NB.c": 0, "NB.b": 2},
        {"NB.c": 0, "NB.b": 3},
        {"NB.c": 0, "NB.b": 4},
        {"NB.c": 0, "NB.b": 5},
        {"NB.c": 1, "NB.b": 0},
        {"NB.c": 1, "NB.b": 1},
        {"NB.c": 1, "NB.b": 2},
        {"NB.c": 1, "NB.b": 3},
        {"NB.c": 1, "NB.b": 4},
        {"NB.c": 1, "NB.b": 5},
    ]


def test_state_connect_combine_left_1():
    """ two 'connected' states,
        the first one has the simplest splitter,
        the second has combiner from the first state
        (i.e. from the Left part of the splitter),
    """
    st1 = State(name="NA", splitter="a")
    st2 = State(name="NB", other_states={"NA": (st1, "b")}, combiner="NA.a")
    assert st2.splitter == "_NA"
    assert st2.splitter_rpn == ["NA.a"]
    assert st2.combiner == st2.left_combiner == st2.left_combiner_all == ["NA.a"]
    assert st2.right_combiner == st2.right_combiner_all == []
    assert st2.splitter_rpn_final == []

    st2.prepare_states(inputs={"NA.a": [3, 5]})
    assert st2.group_for_inputs_final == {}
    assert st2.groups_stack_final == []
    assert st2.states_ind == [{"NA.a": 0}, {"NA.a": 1}]
    assert st2.states_val == [{"NA.a": 3}, {"NA.a": 5}]
    assert st2.states_ind_final == []

    st2.prepare_inputs()
    assert st2.inputs_ind == [{"NB.b": 0}, {"NB.b": 1}]


def test_state_connect_combine_left_2():
    """ two 'connected' states,
        the first one has outer splitter,
        the second has combiner from the first state
        (i.e. from the Left part of the splitter),
    """
    st1 = State(name="NA", splitter=["a", "b"])
    st2 = State(name="NB", other_states={"NA": (st1, "b")}, combiner="NA.a")
    assert st2.splitter == "_NA"
    assert st2.splitter_rpn == ["NA.a", "NA.b", "*"]
    assert st2.combiner == ["NA.a"]
    assert st2.left_combiner_all == st2.left_combiner == ["NA.a"]
    assert st2.right_combiner_all == st2.right_combiner == []
    assert st2.splitter_rpn_final == ["NA.b"]

    st2.prepare_states(inputs={"NA.a": [3, 5], "NA.b": [10, 20]})
    assert st2.group_for_inputs_final == {"NA.b": 0}
    assert st2.groups_stack_final == [[0]]
    assert st2.states_ind == [
        {"NA.a": 0, "NA.b": 0},
        {"NA.a": 0, "NA.b": 1},
        {"NA.a": 1, "NA.b": 0},
        {"NA.a": 1, "NA.b": 1},
    ]
    assert st2.states_val == [
        {"NA.a": 3, "NA.b": 10},
        {"NA.a": 3, "NA.b": 20},
        {"NA.a": 5, "NA.b": 10},
        {"NA.a": 5, "NA.b": 20},
    ]
    assert st2.states_ind_final == [{"NA.b": 0}, {"NA.b": 1}]

    st2.prepare_inputs()
    assert st2.inputs_ind == [{"NB.b": 0}, {"NB.b": 1}, {"NB.b": 2}, {"NB.b": 3}]


def test_state_connect_combine_left_3():
    """ three serially 'connected' states,
        the first one has outer splitter,
        the third one has combiner from the first state
        (i.e. from the Left part of the splitter),
    """
    st1 = State(name="NA", splitter=["a", "b"])
    st2 = State(name="NB", other_states={"NA": (st1, "b")})
    st3 = State(name="NC", other_states={"NB": (st2, "c")}, combiner="NA.a")
    assert st3.splitter == "_NB"
    assert st3.splitter_rpn == ["NA.a", "NA.b", "*"]
    assert st3.combiner == ["NA.a"]
    assert st3.splitter_rpn_final == ["NA.b"]

    st3.prepare_states(inputs={"NA.a": [3, 5], "NA.b": [10, 20]})
    assert st3.group_for_inputs_final == {"NA.b": 0}
    assert st3.groups_stack_final == [[0]]

    assert st3.states_ind == [
        {"NA.a": 0, "NA.b": 0},
        {"NA.a": 0, "NA.b": 1},
        {"NA.a": 1, "NA.b": 0},
        {"NA.a": 1, "NA.b": 1},
    ]
    assert st3.states_val == [
        {"NA.a": 3, "NA.b": 10},
        {"NA.a": 3, "NA.b": 20},
        {"NA.a": 5, "NA.b": 10},
        {"NA.a": 5, "NA.b": 20},
    ]
    assert st3.states_ind_final == [{"NA.b": 0}, {"NA.b": 1}]

    st3.prepare_inputs()
    assert st3.inputs_ind == [{"NC.c": 0}, {"NC.c": 1}, {"NC.c": 2}, {"NC.c": 3}]


def test_state_connect_combine_left_4():
    """ three 'connected' states: testing groups, prepare_states and prepare_inputs,
         the first two states have the simplest splitters,
         the third state has only Left splitter,
         the third state has also combiner from the Left part
      """
    st1 = State(name="NA", splitter="a")
    st2 = State(name="NB", splitter="a")
    st3 = State(
        name="NC",
        splitter=["_NA", "_NB"],
        combiner=["NA.a"],
        other_states={"NA": (st1, "a"), "NB": (st2, "b")},
    )
    assert st3.splitter == ["_NA", "_NB"]
    assert st3.splitter_rpn == ["NA.a", "NB.a", "*"]
    assert st3.splitter_rpn_final == ["NB.a"]
    assert st3.left_combiner_all == st3.left_combiner == st3.combiner == ["NA.a"]
    assert st3.right_combiner_all == st3.right_combiner == []

    st3.prepare_states(inputs={"NA.a": [3, 5], "NB.a": [600, 700]})
    assert st3.group_for_inputs_final == {"NB.a": 0}
    assert st3.groups_stack_final == [[0]]

    assert st3.states_ind == [
        {"NA.a": 0, "NB.a": 0},
        {"NA.a": 0, "NB.a": 1},
        {"NA.a": 1, "NB.a": 0},
        {"NA.a": 1, "NB.a": 1},
    ]
    assert st3.states_val == [
        {"NA.a": 3, "NB.a": 600},
        {"NA.a": 3, "NB.a": 700},
        {"NA.a": 5, "NB.a": 600},
        {"NA.a": 5, "NB.a": 700},
    ]
    assert st3.states_ind_final == [{"NB.a": 0}, {"NB.a": 1}]

    st3.prepare_inputs()
    assert st3.inputs_ind == [
        {"NC.a": 0, "NC.b": 0},
        {"NC.a": 0, "NC.b": 1},
        {"NC.a": 1, "NC.b": 0},
        {"NC.a": 1, "NC.b": 1},
    ]


def test_state_connect_combine_left_5():
    """ three 'connected' states: testing groups, prepare_states and prepare_inputs,
        the first two states have the simplest splitters,
        the third state has scalar splitter in the Left part,
        the third state has also combiner from the Left part
     """
    st1 = State(name="NA", splitter="a")
    st2 = State(name="NB", splitter="a")
    st3 = State(
        name="NC",
        splitter=("_NA", "_NB"),
        combiner=["NA.a"],
        other_states={"NA": (st1, "a"), "NB": (st2, "b")},
    )
    assert st3.splitter == ("_NA", "_NB")
    assert st3.splitter_rpn == ["NA.a", "NB.a", "."]
    st3.set_input_groups()
    assert st3.splitter_rpn_final == []
    assert set(st3.left_combiner_all) == {"NA.a", "NB.a"}
    assert st3.left_combiner == st3.combiner == ["NA.a"]
    assert st3.right_combiner_all == st3.right_combiner == []

    st3.prepare_states(inputs={"NA.a": [3, 5], "NB.a": [600, 700]})
    assert st3.group_for_inputs_final == {}
    assert st3.groups_stack_final == []

    assert st3.states_ind == [{"NA.a": 0, "NB.a": 0}, {"NA.a": 1, "NB.a": 1}]
    assert st3.states_val == [{"NA.a": 3, "NB.a": 600}, {"NA.a": 5, "NB.a": 700}]
    assert st3.states_ind_final == []

    st3.prepare_inputs()
    assert st3.inputs_ind == [{"NC.a": 0, "NC.b": 0}, {"NC.a": 1, "NC.b": 1}]


def test_state_connect_combine_left_6():
    """ two 'connected' states,
        the first one has outer splitter, the second has an additional Right splitter,
        the second also has combiner from the first state
        (i.e. from the Left part of the splitter),
    """
    st1 = State(name="NA", splitter=["a", "b"])
    st2 = State(
        name="NB", splitter="c", other_states={"NA": (st1, "b")}, combiner="NA.a"
    )
    assert st2.splitter == ["_NA", "NB.c"]
    assert st2.splitter_rpn == ["NA.a", "NA.b", "*", "NB.c", "*"]
    assert st2.combiner == ["NA.a"]
    assert st2.left_combiner_all == st2.left_combiner == ["NA.a"]
    assert st2.right_combiner_all == st2.right_combiner == []
    assert st2.splitter_rpn_final == ["NA.b", "NB.c", "*"]

    st2.prepare_states(inputs={"NA.a": [3, 5], "NA.b": [10, 20], "NB.c": [0, 1]})
    assert st2.group_for_inputs_final == {"NA.b": 0, "NB.c": 1}
    assert st2.groups_stack_final == [[0, 1]]

    assert st2.states_ind == [
        {"NA.a": 0, "NA.b": 0, "NB.c": 0},
        {"NA.a": 0, "NA.b": 0, "NB.c": 1},
        {"NA.a": 0, "NA.b": 1, "NB.c": 0},
        {"NA.a": 0, "NA.b": 1, "NB.c": 1},
        {"NA.a": 1, "NA.b": 0, "NB.c": 0},
        {"NA.a": 1, "NA.b": 0, "NB.c": 1},
        {"NA.a": 1, "NA.b": 1, "NB.c": 0},
        {"NA.a": 1, "NA.b": 1, "NB.c": 1},
    ]
    assert st2.states_val == [
        {"NA.a": 3, "NA.b": 10, "NB.c": 0},
        {"NA.a": 3, "NA.b": 10, "NB.c": 1},
        {"NA.a": 3, "NA.b": 20, "NB.c": 0},
        {"NA.a": 3, "NA.b": 20, "NB.c": 1},
        {"NA.a": 5, "NA.b": 10, "NB.c": 0},
        {"NA.a": 5, "NA.b": 10, "NB.c": 1},
        {"NA.a": 5, "NA.b": 20, "NB.c": 0},
        {"NA.a": 5, "NA.b": 20, "NB.c": 1},
    ]
    assert st2.states_ind_final == [
        {"NA.b": 0, "NB.c": 0},
        {"NA.b": 0, "NB.c": 1},
        {"NA.b": 1, "NB.c": 0},
        {"NA.b": 1, "NB.c": 1},
    ]

    st2.prepare_inputs()
    assert st2.inputs_ind == [
        {"NB.b": 0, "NB.c": 0},
        {"NB.b": 0, "NB.c": 1},
        {"NB.b": 1, "NB.c": 0},
        {"NB.b": 1, "NB.c": 1},
        {"NB.b": 2, "NB.c": 0},
        {"NB.b": 2, "NB.c": 1},
        {"NB.b": 3, "NB.c": 0},
        {"NB.b": 3, "NB.c": 1},
    ]


@pytest.mark.parametrize(
    "splitter, other_states, expected_splitter, expected_left, expected_right",
    [
        (None, {"NA": (State(name="NA", splitter="a"), "b")}, "_NA", "_NA", None),
        (
            "b",
            {"NA": (State(name="NA", splitter="a"), "b")},
            ["_NA", "CN.b"],
            "_NA",
            "CN.b",
        ),
        (
            ("b", "c"),
            {"NA": (State(name="NA", splitter="a"), "b")},
            ["_NA", ("CN.b", "CN.c")],
            "_NA",
            ("CN.b", "CN.c"),
        ),
        (
            None,
            {
                "NA": (State(name="NA", splitter="a"), "a"),
                "NB": (State(name="NB", splitter="a"), "b"),
            },
            ["_NA", "_NB"],
            ["_NA", "_NB"],
            None,
        ),
        (
            "b",
            {
                "NA": (State(name="NA", splitter="a"), "a"),
                "NB": (State(name="NB", splitter="a"), "b"),
            },
            [["_NA", "_NB"], "CN.b"],
            ["_NA", "_NB"],
            "CN.b",
        ),
        (
            ["_NA", "b"],
            {
                "NA": (State(name="NA", splitter="a"), "a"),
                "NB": (State(name="NB", splitter="a"), "b"),
            },
            [["_NB", "_NA"], "CN.b"],
            ["_NB", "_NA"],
            "CN.b",
        ),
    ],
)
def test_connect_splitters(
    splitter, other_states, expected_splitter, expected_left, expected_right
):
    st = State(name="CN", splitter=splitter, other_states=other_states)
    st.set_input_groups()
    assert st.splitter == expected_splitter
    assert st.left_splitter == expected_left
    assert st.right_splitter == expected_right


@pytest.mark.parametrize(
    "splitter, other_states",
    [
        (("_NA", "b"), {"NA": (State(name="NA", splitter="a"), "b")}),
        (["b", "_NA"], {"NA": (State(name="NA", splitter="a"), "b")}),
        (
            ["_NB", ["_NA", "b"]],
            {
                "NA": (State(name="NA", splitter="a"), "a"),
                "NB": (State(name="NB", splitter="a"), "b"),
            },
        ),
    ],
)
def test_connect_splitters_exception_1(splitter, other_states):
    with pytest.raises(PydraStateError) as excinfo:
        st = State(name="CN", splitter=splitter, other_states=other_states)
    assert "Left and Right splitters are mixed" in str(excinfo.value)


def test_connect_splitters_exception_2():
    st = State(
        name="CN",
        splitter="_NB",
        other_states={"NA": (State(name="NA", splitter="a"), "b")},
    )
    with pytest.raises(PydraStateError) as excinfo:
        st.set_input_groups()
    assert "can't ask for splitter from NB" in str(excinfo.value)


def test_connect_splitters_exception_3():
    with pytest.raises(PydraStateError) as excinfo:
        st = State(
            name="CN",
            splitter="_NB",
            other_states=["NA", (State(name="NA", splitter="a"), "b")],
        )
    assert "other states has to be a dictionary" == str(excinfo.value)



import shutil


def pytest_addoption(parser):
    parser.addoption("--dask", action="store_true", help="run all combinations")


def pytest_generate_tests(metafunc):
    if "plugin_dask_opt" in metafunc.fixturenames:
        if bool(shutil.which("sbatch")):
            Plugins = ["slurm"]
        else:
            Plugins = ["cf"]
        if metafunc.config.getoption("dask"):
            Plugins.append("dask")
        metafunc.parametrize("plugin_dask_opt", Plugins)

    if "plugin" in metafunc.fixturenames:
        if metafunc.config.getoption("dask"):
            Plugins = []
        elif bool(shutil.which("sbatch")):
            Plugins = ["slurm"]
        else:
            Plugins = ["cf"]
        metafunc.parametrize("plugin", Plugins)



import os, shutil
import subprocess as sp
from pathlib import Path
import pytest

from ..core import Workflow
from ..task import ShellCommandTask
from ..submitter import Submitter
from ..boutiques import BoshTask
from .utils import result_no_submitter, result_submitter, no_win

need_bosh_docker = pytest.mark.skipif(
    shutil.which("docker") is None
    or sp.call(["docker", "info"])
    or sp.call(["which", "bosh"]),
    reason="requires docker and bosh",
)

Infile = Path(__file__).resolve().parent / "data_tests" / "test.nii.gz"


@no_win
@need_bosh_docker
@pytest.mark.flaky(reruns=3)  # need for travis
@pytest.mark.parametrize(
    "maskfile", ["test_brain.nii.gz", "test_brain", "test_brain.nii"]
)
@pytest.mark.parametrize("results_function", [result_no_submitter, result_submitter])
def test_boutiques_1(maskfile, plugin, results_function):
    """ simple task to run fsl.bet using BoshTask"""
    btask = BoshTask(name="NA", zenodo_id="1482743")
    btask.inputs.infile = Infile
    btask.inputs.maskfile = maskfile
    res = results_function(btask, plugin)

    assert res.output.return_code == 0

    # checking if the outfile exists and if it has proper name
    assert res.output.outfile.name == "test_brain.nii.gz"
    assert res.output.outfile.exists()
    # other files should also have proper names, but they do not exist
    assert res.output.out_outskin_off.name == "test_brain_outskin_mesh.off"
    assert not res.output.out_outskin_off.exists()


@no_win
@need_bosh_docker
@pytest.mark.flaky(reruns=3)
def test_boutiques_spec_1():
    """ testing spec: providing input/output fields names"""
    btask = BoshTask(
        name="NA",
        zenodo_id="1482743",
        infile=Infile,
        maskfile="test_brain.nii.gz",
        input_spec_names=["infile", "maskfile"],
        output_spec_names=["outfile", "out_outskin_off"],
    )

    assert len(btask.input_spec.fields) == 2
    assert btask.input_spec.fields[0][0] == "infile"
    assert btask.input_spec.fields[1][0] == "maskfile"
    assert hasattr(btask.inputs, "infile")
    assert hasattr(btask.inputs, "maskfile")

    assert len(btask.output_spec.fields) == 2
    assert btask.output_spec.fields[0][0] == "outfile"
    assert btask.output_spec.fields[1][0] == "out_outskin_off"


@no_win
@need_bosh_docker
@pytest.mark.flaky(reruns=3)
def test_boutiques_spec_2():
    """ testing spec: providing partial input/output fields names"""
    btask = BoshTask(
        name="NA",
        zenodo_id="1482743",
        infile=Infile,
        maskfile="test_brain.nii.gz",
        input_spec_names=["infile"],
        output_spec_names=[],
    )

    assert len(btask.input_spec.fields) == 1
    assert btask.input_spec.fields[0][0] == "infile"
    assert hasattr(btask.inputs, "infile")
    # input doesn't see maskfile
    assert not hasattr(btask.inputs, "maskfile")

    assert len(btask.output_spec.fields) == 0


@no_win
@need_bosh_docker
@pytest.mark.flaky(reruns=3)
@pytest.mark.parametrize(
    "maskfile", ["test_brain.nii.gz", "test_brain", "test_brain.nii"]
)
def test_boutiques_wf_1(maskfile, plugin):
    """ wf with one task that runs fsl.bet using BoshTask"""
    wf = Workflow(name="wf", input_spec=["maskfile", "infile"])
    wf.inputs.maskfile = maskfile
    wf.inputs.infile = Infile

    wf.add(
        BoshTask(
            name="bet",
            zenodo_id="1482743",
            infile=wf.lzin.infile,
            maskfile=wf.lzin.maskfile,
        )
    )

    wf.set_output([("outfile", wf.bet.lzout.outfile)])

    with Submitter(plugin=plugin) as sub:
        wf(submitter=sub)

    res = wf.result()
    assert res.output.outfile.name == "test_brain.nii.gz"
    assert res.output.outfile.exists()


@no_win
@need_bosh_docker
@pytest.mark.flaky(reruns=3)
@pytest.mark.parametrize(
    "maskfile", ["test_brain.nii.gz", "test_brain", "test_brain.nii"]
)
def test_boutiques_wf_2(maskfile, plugin):
    """ wf with two BoshTasks (fsl.bet and fsl.stats) and one ShellTask"""
    wf = Workflow(name="wf", input_spec=["maskfile", "infile"])
    wf.inputs.maskfile = maskfile
    wf.inputs.infile = Infile

    wf.add(
        BoshTask(
            name="bet",
            zenodo_id="1482743",
            infile=wf.lzin.infile,
            maskfile=wf.lzin.maskfile,
        )
    )
    wf.add(
        BoshTask(
            name="stat", zenodo_id="3240521", input_file=wf.bet.lzout.outfile, v=True
        )
    )
    wf.add(ShellCommandTask(name="cat", executable="cat", args=wf.stat.lzout.output))

    wf.set_output(
        [
            ("outfile_bet", wf.bet.lzout.outfile),
            ("out_stat", wf.stat.lzout.output),
            ("out", wf.cat.lzout.stdout),
        ]
    )

    with Submitter(plugin=plugin) as sub:
        wf(submitter=sub)

    res = wf.result()
    assert res.output.outfile_bet.name == "test_brain.nii.gz"
    assert res.output.outfile_bet.exists()

    assert res.output.out_stat.name == "output.txt"
    assert res.output.out_stat.exists()

    assert int(res.output.out.rstrip().split()[0]) == 11534336
    assert float(res.output.out.rstrip().split()[1]) == 11534336.0



# -*- coding: utf-8 -*-

import typing as ty
import os, sys
import pytest

from ... import mark
from ..task import AuditFlag, ShellCommandTask, DockerTask, SingularityTask
from ...utils.messenger import FileMessenger, PrintMessenger, collect_messages
from .utils import gen_basic_wf

no_win = pytest.mark.skipif(
    sys.platform.startswith("win"),
    reason="docker/singularity command not adjusted for windows",
)


@mark.task
def funaddtwo(a):
    return a + 2


def test_output():
    nn = funaddtwo(a=3)
    res = nn._run()
    assert res.output.out == 5


def test_name_conflict():
    """ raise error if task name conflicts with a class attribute or method"""
    with pytest.raises(ValueError) as excinfo1:
        nn = funaddtwo(name="split", a=3)
    assert "Cannot use names of attributes or methods" in str(excinfo1.value)
    with pytest.raises(ValueError) as excinfo2:
        nn = funaddtwo(name="checksum", a=3)
    assert "Cannot use names of attributes or methods" in str(excinfo2.value)


def test_numpy():
    """ checking if mark.task works for numpy functions"""
    np = pytest.importorskip("numpy")
    fft = mark.annotate({"a": np.ndarray, "return": float})(np.fft.fft)
    fft = mark.task(fft)()
    arr = np.array([[1, 10], [2, 20]])
    fft.inputs.a = arr
    res = fft()
    assert np.allclose(np.fft.fft(arr), res.output.out)


@pytest.mark.xfail(reason="cp.dumps(func) depends on the system/setup, TODO!!")
def test_checksum():
    nn = funaddtwo(a=3)
    assert (
        nn.checksum
        == "FunctionTask_abb4e7cc03b13d0e73884b87d142ed5deae6a312275187a9d8df54407317d7d3"
    )


def test_annotated_func():
    @mark.task
    def testfunc(
        a: int, b: float = 0.1
    ) -> ty.NamedTuple("Output", [("out_out", float)]):
        return a + b

    funky = testfunc(a=1)
    assert hasattr(funky.inputs, "a")
    assert hasattr(funky.inputs, "b")
    assert hasattr(funky.inputs, "_func")
    assert getattr(funky.inputs, "a") == 1
    assert getattr(funky.inputs, "b") == 0.1
    assert getattr(funky.inputs, "_func") is not None
    assert set(funky.output_names) == set(["out_out"])
    # assert funky.inputs.hash == '17772c3aec9540a8dd3e187eecd2301a09c9a25c6e371ddd86e31e3a1ecfeefa'
    assert funky.__class__.__name__ + "_" + funky.inputs.hash == funky.checksum

    result = funky()
    assert hasattr(result, "output")
    assert hasattr(result.output, "out_out")
    assert result.output.out_out == 1.1

    assert os.path.exists(funky.cache_dir / funky.checksum / "_result.pklz")
    funky.result()  # should not recompute
    funky.inputs.a = 2
    # assert funky.checksum == '537d25885fd2ea5662b7701ba02c132c52a9078a3a2d56aa903a777ea90e5536'
    assert funky.result() is None
    funky()
    result = funky.result()
    assert result.output.out_out == 2.1

    help = funky.help(returnhelp=True)
    assert help == [
        "Help for FunctionTask",
        "Input Parameters:",
        "- a: int",
        "- b: float (default: 0.1)",
        "- _func: str",
        "Output Parameters:",
        "- out_out: float",
    ]


def test_annotated_func_multreturn():
    """ the function has two elements in the return statement"""

    @mark.task
    def testfunc(
        a: float,
    ) -> ty.NamedTuple("Output", [("fractional", float), ("integer", int)]):
        import math

        return math.modf(a)

    funky = testfunc(a=3.5)
    assert hasattr(funky.inputs, "a")
    assert hasattr(funky.inputs, "_func")
    assert getattr(funky.inputs, "a") == 3.5
    assert getattr(funky.inputs, "_func") is not None
    assert set(funky.output_names) == set(["fractional", "integer"])
    assert funky.__class__.__name__ + "_" + funky.inputs.hash == funky.checksum

    result = funky()
    assert os.path.exists(funky.cache_dir / funky.checksum / "_result.pklz")
    assert hasattr(result, "output")
    assert hasattr(result.output, "fractional")
    assert result.output.fractional == 0.5
    assert hasattr(result.output, "integer")
    assert result.output.integer == 3

    help = funky.help(returnhelp=True)
    assert help == [
        "Help for FunctionTask",
        "Input Parameters:",
        "- a: float",
        "- _func: str",
        "Output Parameters:",
        "- fractional: float",
        "- integer: int",
    ]


def test_annotated_func_multreturn_exception():
    """function has two elements in the return statement,
        but three element provided in the spec - should raise an error
    """

    @mark.task
    def testfunc(
        a: float,
    ) -> ty.NamedTuple(
        "Output", [("fractional", float), ("integer", int), ("whoknows", int)]
    ):
        import math

        return math.modf(a)

    funky = testfunc(a=3.5)
    with pytest.raises(Exception) as excinfo:
        funky()
    assert "expected 3 elements" in str(excinfo.value)


def test_halfannotated_func():
    @mark.task
    def testfunc(a, b) -> int:
        return a + b

    funky = testfunc(a=10, b=20)
    assert hasattr(funky.inputs, "a")
    assert hasattr(funky.inputs, "b")
    assert hasattr(funky.inputs, "_func")
    assert getattr(funky.inputs, "a") == 10
    assert getattr(funky.inputs, "b") == 20
    assert getattr(funky.inputs, "_func") is not None
    assert set(funky.output_names) == set(["out"])
    assert funky.__class__.__name__ + "_" + funky.inputs.hash == funky.checksum

    result = funky()
    assert hasattr(result, "output")
    assert hasattr(result.output, "out")
    assert result.output.out == 30

    assert os.path.exists(funky.cache_dir / funky.checksum / "_result.pklz")

    funky.result()  # should not recompute
    funky.inputs.a = 11
    assert funky.result() is None
    funky()
    result = funky.result()
    assert result.output.out == 31
    help = funky.help(returnhelp=True)

    assert help == [
        "Help for FunctionTask",
        "Input Parameters:",
        "- a: _empty",
        "- b: _empty",
        "- _func: str",
        "Output Parameters:",
        "- out: int",
    ]


def test_halfannotated_func_multreturn():
    @mark.task
    def testfunc(a, b) -> (int, int):
        return a + 1, b + 1

    funky = testfunc(a=10, b=20)
    assert hasattr(funky.inputs, "a")
    assert hasattr(funky.inputs, "b")
    assert hasattr(funky.inputs, "_func")
    assert getattr(funky.inputs, "a") == 10
    assert getattr(funky.inputs, "b") == 20
    assert getattr(funky.inputs, "_func") is not None
    assert set(funky.output_names) == set(["out1", "out2"])
    assert funky.__class__.__name__ + "_" + funky.inputs.hash == funky.checksum

    result = funky()
    assert hasattr(result, "output")
    assert hasattr(result.output, "out1")
    assert result.output.out1 == 11

    assert os.path.exists(funky.cache_dir / funky.checksum / "_result.pklz")

    funky.result()  # should not recompute
    funky.inputs.a = 11
    assert funky.result() is None
    funky()
    result = funky.result()
    assert result.output.out1 == 12
    help = funky.help(returnhelp=True)

    assert help == [
        "Help for FunctionTask",
        "Input Parameters:",
        "- a: _empty",
        "- b: _empty",
        "- _func: str",
        "Output Parameters:",
        "- out1: int",
        "- out2: int",
    ]


def test_notannotated_func():
    @mark.task
    def no_annots(c, d):
        return c + d

    natask = no_annots(c=17, d=3.2)
    assert hasattr(natask.inputs, "c")
    assert hasattr(natask.inputs, "d")
    assert hasattr(natask.inputs, "_func")

    result = natask._run()
    assert hasattr(result, "output")
    assert hasattr(result.output, "out")
    assert result.output.out == 20.2


def test_notannotated_func_returnlist():
    @mark.task
    def no_annots(c, d):
        return [c, d]

    natask = no_annots(c=17, d=3.2)
    result = natask._run()
    assert hasattr(result.output, "out")
    assert result.output.out == [17, 3.2]


def test_halfannotated_func_multrun_returnlist():
    @mark.task
    def no_annots(c, d) -> (list, float):
        return [c, d], c + d

    natask = no_annots(c=17, d=3.2)
    result = natask._run()

    assert hasattr(result.output, "out1")
    assert hasattr(result.output, "out2")
    assert result.output.out1 == [17, 3.2]
    assert result.output.out2 == 20.2


def test_notannotated_func_multreturn():
    """ no annotation and multiple values are returned
        all elements should be returned as a tuple ans set to "out"
    """

    @mark.task
    def no_annots(c, d):
        return c + d, c - d

    natask = no_annots(c=17, d=3.2)
    assert hasattr(natask.inputs, "c")
    assert hasattr(natask.inputs, "d")
    assert hasattr(natask.inputs, "_func")

    result = natask._run()
    assert hasattr(result, "output")
    assert hasattr(result.output, "out")
    assert result.output.out == (20.2, 13.8)


def test_exception_func():
    @mark.task
    def raise_exception(c, d):
        raise Exception()

    bad_funk = raise_exception(c=17, d=3.2)
    assert pytest.raises(Exception, bad_funk)


def test_result_none_1():
    """ checking if None is properly returned as the result"""

    @mark.task
    def fun_none(x):
        return None

    task = fun_none(name="none", x=3)
    res = task()
    assert res.output.out is None


def test_result_none_2():
    """ checking if None is properly set for all outputs """

    @mark.task
    def fun_none(x) -> (ty.Any, ty.Any):
        return None

    task = fun_none(name="none", x=3)
    res = task()
    assert res.output.out1 is None
    assert res.output.out2 is None


def test_audit_prov(tmpdir):
    @mark.task
    def testfunc(a: int, b: float = 0.1) -> ty.NamedTuple("Output", [("out", float)]):
        return a + b

    # printing the audit message
    funky = testfunc(a=1, audit_flags=AuditFlag.PROV, messengers=PrintMessenger())
    funky.cache_dir = tmpdir
    funky()

    # saving the audit message into the file
    funky = testfunc(a=2, audit_flags=AuditFlag.PROV, messengers=FileMessenger())
    message_path = tmpdir / funky.checksum / "messages"
    funky.cache_dir = tmpdir
    funky.messenger_args = dict(message_dir=message_path)
    funky()

    collect_messages(tmpdir / funky.checksum, message_path, ld_op="compact")
    assert (tmpdir / funky.checksum / "messages.jsonld").exists()


def test_audit_all(tmpdir):
    @mark.task
    def testfunc(a: int, b: float = 0.1) -> ty.NamedTuple("Output", [("out", float)]):
        return a + b

    funky = testfunc(a=2, audit_flags=AuditFlag.ALL, messengers=FileMessenger())
    message_path = tmpdir / funky.checksum / "messages"
    funky.cache_dir = tmpdir
    funky.messenger_args = dict(message_dir=message_path)
    funky()
    from glob import glob

    assert len(glob(str(tmpdir / funky.checksum / "proc*.log"))) == 1
    assert len(glob(str(message_path / "*.jsonld"))) == 6

    # commented out to speed up testing
    collect_messages(tmpdir / funky.checksum, message_path, ld_op="compact")
    assert (tmpdir / funky.checksum / "messages.jsonld").exists()


@no_win
def test_shell_cmd(tmpdir):
    cmd = ["echo", "hail", "pydra"]

    # all args given as executable
    shelly = ShellCommandTask(name="shelly", executable=cmd)
    assert shelly.cmdline == " ".join(cmd)
    res = shelly._run()
    assert res.output.stdout == " ".join(cmd[1:]) + "\n"

    # separate command into exec + args
    shelly = ShellCommandTask(executable=cmd[0], args=cmd[1:])
    assert shelly.inputs.executable == "echo"
    assert shelly.cmdline == " ".join(cmd)
    res = shelly._run()
    assert res.output.return_code == 0
    assert res.output.stdout == " ".join(cmd[1:]) + "\n"


def test_container_cmds(tmpdir):
    containy = DockerTask(name="containy", executable="pwd")
    with pytest.raises(AttributeError) as excinfo:
        containy.cmdline
    assert "mandatory" in str(excinfo.value)
    containy.inputs.image = "busybox"
    assert containy.cmdline


@no_win
def test_docker_cmd(tmpdir):
    docky = DockerTask(name="docky", executable="pwd", image="busybox")
    assert (
        docky.cmdline
        == f"docker run --rm -v {docky.output_dir}:/output_pydra:rw -w /output_pydra busybox pwd"
    )
    docky.inputs.container_xargs = ["--rm -it"]
    assert (
        docky.cmdline
        == f"docker run --rm -it -v {docky.output_dir}:/output_pydra:rw -w /output_pydra busybox pwd"
    )
    docky.inputs.bindings = [
        ("/local/path", "/container/path", "ro"),
        ("/local2", "/container2", None),
    ]
    assert docky.cmdline == (
        "docker run --rm -it -v /local/path:/container/path:ro"
        f" -v /local2:/container2:rw -v {docky.output_dir}:/output_pydra:rw -w /output_pydra busybox pwd"
    )


@no_win
def test_singularity_cmd(tmpdir):
    # todo how this should be done?
    image = "library://sylabsed/linux/alpine"
    singu = SingularityTask(name="singi", executable="pwd", image=image)
    assert (
        singu.cmdline
        == f"singularity exec -B {singu.output_dir}:/output_pydra:rw {image} pwd"
    )
    singu.inputs.bindings = [
        ("/local/path", "/container/path", "ro"),
        ("/local2", "/container2", None),
    ]
    assert singu.cmdline == (
        "singularity exec -B /local/path:/container/path:ro"
        f" -B /local2:/container2:rw -B {singu.output_dir}:/output_pydra:rw {image} pwd"
    )


def test_functask_callable(tmpdir):
    # no submitter or plugin
    foo = funaddtwo(a=1)
    res = foo()
    assert res.output.out == 3
    assert foo.plugin is None

    # plugin
    bar = funaddtwo(a=2)
    res = bar(plugin="cf")
    assert res.output.out == 4
    assert bar.plugin is None

    foo2 = funaddtwo(a=3)
    foo2.plugin = "cf"
    res = foo2()
    assert res.output.out == 5
    assert foo2.plugin == "cf"


def test_taskhooks(tmpdir, capsys):
    foo = funaddtwo(name="foo", a=1, cache_dir=tmpdir)
    assert foo.hooks
    # ensure all hooks are defined
    for attr in ("pre_run", "post_run", "pre_run_task", "post_run_task"):
        hook = getattr(foo.hooks, attr)
        assert hook() is None

    def myhook(task, *args):
        print("I was called")

    foo.hooks.pre_run = myhook
    foo()
    captured = capsys.readouterr()
    assert "I was called\n" in captured.out
    del captured

    # setting unknown hook should not be allowed
    with pytest.raises(AttributeError):
        foo.hooks.mid_run = myhook

    # set all hooks
    foo.hooks.post_run = myhook
    foo.hooks.pre_run_task = myhook
    foo.hooks.post_run_task = myhook
    foo.inputs.a = 2  # ensure not pre-cached
    foo()
    captured = capsys.readouterr()
    assert captured.out.count("I was called\n") == 4
    del captured

    # hooks are independent across tasks by default
    bar = funaddtwo(name="bar", a=3, cache_dir=tmpdir)
    assert bar.hooks is not foo.hooks
    # but can be shared across tasks
    bar.hooks = foo.hooks
    # and workflows
    wf = gen_basic_wf()
    wf.tmpdir = tmpdir
    wf.hooks = bar.hooks
    assert foo.hooks == bar.hooks == wf.hooks

    wf(plugin="cf")
    captured = capsys.readouterr()
    assert captured.out.count("I was called\n") == 4
    del captured

    # reset all hooks
    foo.hooks.reset()
    for attr in ("pre_run", "post_run", "pre_run_task", "post_run_task"):
        hook = getattr(foo.hooks, attr)
        assert hook() is None



# -*- coding: utf-8 -*-

import pytest
import random

from ..functions import task, annotate
from ...engine.task import FunctionTask


def test_task_equivalence():
    def add_two(a):
        return a + 2

    canonical = FunctionTask(add_two, a=3)

    decorated1 = task(add_two)(a=3)

    @task
    def addtwo(a):
        return a + 2

    decorated2 = addtwo(a=3)

    assert canonical.checksum == decorated1.checksum

    c_res = canonical._run()
    d1_res = decorated1._run()
    d2_res = decorated2._run()

    assert c_res.output.hash == d1_res.output.hash
    assert c_res.output.hash == d2_res.output.hash


def test_annotation_equivalence():
    def direct(a: int) -> int:
        return a + 2

    @annotate({"return": int})
    def partial(a: int):
        return a + 2

    @annotate({"a": int, "return": int})
    def indirect(a):
        return a + 2

    assert direct.__annotations__ == partial.__annotations__
    assert direct.__annotations__ == indirect.__annotations__

    # Run functions to ensure behavior is unaffected
    a = random.randint(0, (1 << 32) - 3)
    assert direct(a) == partial(a)
    assert direct(a) == indirect(a)


def test_annotation_override():
    @annotate({"a": float, "return": float})
    def annotated(a: int) -> int:
        return a + 2

    assert annotated.__annotations__ == {"a": float, "return": float}


def test_invalid_annotation():
    with pytest.raises(TypeError):

        @annotate({"b": int})
        def addtwo(a):
            return a + 2


def test_annotated_task():
    @task
    def square(in_val: float):
        return in_val ** 2

    res = square(in_val=2.0)()
    assert res.output.out == 4.0


def test_return_annotated_task():
    @task
    @annotate({"in_val": float, "return": {"squared": float}})
    def square(in_val):
        return in_val ** 2

    res = square(in_val=2.0)()
    assert res.output.squared == 4.0


def test_return_halfannotated_annotated_task():
    @task
    @annotate({"in_val": float, "return": float})
    def square(in_val):
        return in_val ** 2

    res = square(in_val=2.0)()
    assert res.output.out == 4.0


def test_return_annotated_task_multiple_output():
    @task
    @annotate({"in_val": float, "return": {"squared": float, "cubed": float}})
    def square(in_val):
        return in_val ** 2, in_val ** 3

    res = square(in_val=2.0)()
    assert res.output.squared == 4.0
    assert res.output.cubed == 8.0


def test_return_halfannotated_task_multiple_output():
    @task
    @annotate({"in_val": float, "return": (float, float)})
    def square(in_val):
        return in_val ** 2, in_val ** 3

    res = square(in_val=2.0)()
    assert res.output.out1 == 4.0
    assert res.output.out2 == 8.0



import typing as ty
import json
import attr
from urllib.request import urlretrieve
from pathlib import Path
from functools import reduce

from ..utils.messenger import AuditFlag
from ..engine import ShellCommandTask
from ..engine.specs import SpecInfo, ShellSpec, ShellOutSpec, File, attr_fields
from .helpers_file import is_local_file


class BoshTask(ShellCommandTask):
    """Shell Command Task based on the Boutiques descriptor"""

    def __init__(
        self,
        zenodo_id=None,
        bosh_file=None,
        audit_flags: AuditFlag = AuditFlag.NONE,
        cache_dir=None,
        input_spec_names: ty.Optional[ty.List] = None,
        messenger_args=None,
        messengers=None,
        name=None,
        output_spec_names: ty.Optional[ty.List] = None,
        rerun=False,
        strip=False,
        **kwargs,
    ):
        """
        Initialize this task.

        Parameters
        ----------
        zenodo_id: :obj: str
            Zenodo ID
        bosh_file : : str
            json file with the boutiques descriptors
        audit_flags : :obj:`pydra.utils.messenger.AuditFlag`
            Auditing configuration
        cache_dir : :obj:`os.pathlike`
            Cache directory
        input_spec_names : :obj: list
            Input names for input_spec.
        messenger_args :
            TODO
        messengers :
            TODO
        name : :obj:`str`
            Name of this task.
        output_spec_names : :obj: list
            Output names for output_spec.
        strip : :obj:`bool`
            TODO

        """
        self.cache_dir = cache_dir
        if (bosh_file and zenodo_id) or not (bosh_file or zenodo_id):
            raise Exception("either bosh or zenodo_id has to be specified")
        elif zenodo_id:
            self.bosh_file = self._download_spec(zenodo_id)
        else:  # bosh_file
            self.bosh_file = bosh_file

        with self.bosh_file.open() as f:
            self.bosh_spec = json.load(f)

        self.input_spec = self._prepare_input_spec(names_subset=input_spec_names)
        self.output_spec = self._prepare_output_spec(names_subset=output_spec_names)
        self.bindings = ["-v", f"{self.bosh_file.parent}:{self.bosh_file.parent}:ro"]

        super(BoshTask, self).__init__(
            name=name,
            input_spec=self.input_spec,
            output_spec=self.output_spec,
            executable=["bosh", "exec", "launch"],
            args=["-s"],
            audit_flags=audit_flags,
            messengers=messengers,
            messenger_args=messenger_args,
            cache_dir=self.cache_dir,
            strip=strip,
            rerun=rerun,
            **kwargs,
        )
        self.strip = strip

    def _download_spec(self, zenodo_id):
        """
        usind boutiques Searcher to find url of zenodo file for a specific id,
        and download the file to self.cache_dir
        """
        from boutiques.searcher import Searcher

        searcher = Searcher(zenodo_id, exact_match=True)
        hits = searcher.zenodo_search().json()["hits"]["hits"]
        if len(hits) == 0:
            raise Exception(f"can't find zenodo spec for {zenodo_id}")
        elif len(hits) > 1:
            raise Exception(f"too many hits for {zenodo_id}")
        else:
            zenodo_url = hits[0]["files"][0]["links"]["self"]
            zenodo_file = self.cache_dir / f"zenodo.{zenodo_id}.json"
            urlretrieve(zenodo_url, zenodo_file)
            return zenodo_file

    def _prepare_input_spec(self, names_subset=None):
        """ creating input spec from the zenodo file
            if name_subset provided, only names from the subset will be used in the spec
        """
        binputs = self.bosh_spec["inputs"]
        self._input_spec_keys = {}
        fields = []
        for input in binputs:
            name = input["id"]
            if names_subset is None:
                pass
            elif name not in names_subset:
                continue
            else:
                names_subset.remove(name)
            if input["type"] == "File":
                tp = File
            elif input["type"] == "String":
                tp = str
            elif input["type"] == "Number":
                tp = float
            elif input["type"] == "Flag":
                tp = bool
            else:
                tp = None
            # adding list
            if tp and "list" in input and input["list"]:
                tp = ty.List[tp]

            mdata = {
                "help_string": input.get("description", None) or input["name"],
                "mandatory": not input["optional"],
                "argstr": input.get("command-line-flag", None),
            }
            fields.append((name, tp, mdata))
            self._input_spec_keys[input["value-key"]] = "{" + f"{name}" + "}"
        if names_subset:
            raise RuntimeError(f"{names_subset} are not in the zenodo input spec")
        spec = SpecInfo(name="Inputs", fields=fields, bases=(ShellSpec,))
        return spec

    def _prepare_output_spec(self, names_subset=None):
        """ creating output spec from the zenodo file
            if name_subset provided, only names from the subset will be used in the spec
        """
        boutputs = self.bosh_spec["output-files"]
        fields = []
        for output in boutputs:
            name = output["id"]
            if names_subset is None:
                pass
            elif name not in names_subset:
                continue
            else:
                names_subset.remove(name)
            path_template = reduce(
                lambda s, r: s.replace(*r),
                self._input_spec_keys.items(),
                output["path-template"],
            )
            mdata = {
                "help_string": output.get("description", None) or output["name"],
                "mandatory": not output["optional"],
                "output_file_template": path_template,
            }
            fields.append((name, attr.ib(type=File, metadata=mdata)))

        if names_subset:
            raise RuntimeError(f"{names_subset} are not in the zenodo output spec")
        spec = SpecInfo(name="Outputs", fields=fields, bases=(ShellOutSpec,))
        return spec

    def _command_args_single(self, state_ind, ind=None):
        """Get command line arguments for a single state"""
        input_filepath = self._bosh_invocation_file(state_ind=state_ind, ind=ind)
        cmd_list = (
            self.inputs.executable
            + [str(self.bosh_file), input_filepath]
            + self.inputs.args
            + self.bindings
        )
        return cmd_list

    def _bosh_invocation_file(self, state_ind, ind=None):
        """creating bosh invocation file - json file with inputs values"""
        input_json = {}
        for f in attr_fields(self.inputs):
            if f.name in ["executable", "args"]:
                continue
            if self.state and f"{self.name}.{f.name}" in state_ind:
                value = getattr(self.inputs, f.name)[state_ind[f"{self.name}.{f.name}"]]
            else:
                value = getattr(self.inputs, f.name)
            # adding to the json file if specified by the user
            if value is not attr.NOTHING and value != "NOTHING":
                if is_local_file(f):
                    value = Path(value)
                    self.bindings.extend(["-v", f"{value.parent}:{value.parent}:ro"])
                    value = str(value)

                input_json[f.name] = value

        filename = self.cache_dir / f"{self.name}-{ind}.json"
        with open(filename, "w") as jsonfile:
            json.dump(input_json, jsonfile)

        return str(filename)
